# -*- coding: utf-8 -*-
"""Date parsing utilities"""
from calendar import monthrange
from datetime import datetime
from parser import ParserError
from typing import Optional, Dict, Tuple

import dateutil
from dateutil.parser import _timelex, parserinfo

from hdx.utilities import raisefrom

default_sd_year = 1
default_date = datetime(year=default_sd_year, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
default_ed_year = 9990
default_enddate = datetime(year=default_ed_year, month=12, day=31, hour=0, minute=0, second=0, microsecond=0)


# Ugly copy and paste from dateutil.parser._parser._ymd with dayfirst modified to mean day is outer value
# ie. dayfirst prefers dmy or ymd where in dateutil it prefers dmy and ydm!
class _ymd(list):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.century_specified = False
        self.dstridx = None
        self.mstridx = None
        self.ystridx = None

    @property
    def has_year(self):
        return self.ystridx is not None

    @property
    def has_month(self):
        return self.mstridx is not None

    @property
    def has_day(self):
        return self.dstridx is not None

    def could_be_day(self, value):
        if self.has_day:
            return False
        elif not self.has_month:
            return 1 <= value <= 31
        elif not self.has_year:
            # Be permissive, assume leap year
            month = self[self.mstridx]
            return 1 <= value <= monthrange(2000, month)[1]
        else:
            month = self[self.mstridx]
            year = self[self.ystridx]
            return 1 <= value <= monthrange(year, month)[1]

    def append(self, val, label=None):
        if hasattr(val, '__len__'):
            if val.isdigit() and len(val) > 2:
                self.century_specified = True
                if label not in [None, 'Y']:  # pragma: no cover
                    raise ValueError(label)
                label = 'Y'
        elif val > 100:
            self.century_specified = True
            if label not in [None, 'Y']:  # pragma: no cover
                raise ValueError(label)
            label = 'Y'

        super(self.__class__, self).append(int(val))

        if label == 'M':
            if self.has_month:
                raise ValueError('Month is already set')
            self.mstridx = len(self) - 1
        elif label == 'D':
            if self.has_day:
                raise ValueError('Day is already set')
            self.dstridx = len(self) - 1
        elif label == 'Y':
            if self.has_year:
                raise ValueError('Year is already set')
            self.ystridx = len(self) - 1

    def _resolve_from_stridxs(self, strids):
        """
        Try to resolve the identities of year/month/day elements using
        ystridx, mstridx, and dstridx, if enough of these are specified.
        """
        if len(self) == 3 and len(strids) == 2:
            # we can back out the remaining stridx value
            missing = [x for x in range(3) if x not in strids.values()]
            key = [x for x in ['y', 'm', 'd'] if x not in strids]
            assert len(missing) == len(key) == 1
            key = key[0]
            val = missing[0]
            strids[key] = val

        assert len(self) == len(strids)  # otherwise this should not be called
        out = {key: self[strids[key]] for key in strids}
        return (out.get('y'), out.get('m'), out.get('d'))

    def resolve_ymd(self, yearfirst, dayfirst):
        len_ymd = len(self)
        year, month, day = (None, None, None)

        strids = (('y', self.ystridx),
                  ('m', self.mstridx),
                  ('d', self.dstridx))

        strids = {key: val for key, val in strids if val is not None}
        if (len(self) == len(strids) > 0 or
                (len(self) == 3 and len(strids) == 2)):
            return self._resolve_from_stridxs(strids)

        mstridx = self.mstridx

        if len_ymd > 3:
            raise ValueError("More than three YMD values")
        elif len_ymd == 1 or (mstridx is not None and len_ymd == 2):
            # One member, or two members with a month string
            if mstridx is not None:
                month = self[mstridx]
                # since mstridx is 0 or 1, self[mstridx-1] always
                # looks up the other element
                other = self[mstridx - 1]
            else:
                other = self[0]

            if len_ymd > 1 or mstridx is None:
                if other > 31:
                    year = other
                else:
                    day = other

        elif len_ymd == 2:
            # Two members with numbers
            if self[0] > 31:
                # 99-01
                year, month = self
            elif self[1] > 31:
                # 01-99
                month, year = self
            elif dayfirst and self[1] <= 12:
                # 13-01
                day, month = self
            else:
                # 01-13
                month, day = self

        elif len_ymd == 3:
            # Three members
            if mstridx == 0:
                if self[1] > 31:
                    # Apr-2003-25
                    month, year, day = self
                else:
                    month, day, year = self
            elif mstridx == 1:
                if self[0] > 31 or (yearfirst and self[2] <= 31):
                    # 99-Jan-01
                    year, month, day = self
                else:
                    # 01-Jan-01
                    # Give precedence to day-first, since
                    # two-digit years is usually hand-written.
                    day, month, year = self

            elif mstridx == 2:
                # WTF!?
                if self[1] > 31:
                    # 01-99-Jan
                    day, year, month = self
                else:
                    # 99-01-Jan
                    year, day, month = self

            else:
                if (self[0] > 31 or
                        self.ystridx == 0 or
                        (yearfirst and self[1] <= 12 and self[2] <= 31)):
                    # 99-01-01
                    if dayfirst and self[1] <= 12:  # CHANGED
                        year, month, day = self
                    else:
                        year, day, month = self
                elif self[0] > 12 or (dayfirst and self[1] <= 12):
                    # 13-01-01
                    day, month, year = self
                else:
                    # 01-13-01
                    month, day, year = self

        return year, month, day


# Ugly copy and paste from dateutil.dateparser with minor changes
class DateParser(dateutil.parser.parser):  # pragma: no cover
    def _parse(self, timestr, dayfirst=None, yearfirst=None, fuzzy=False,
               fuzzy_with_tokens=False):
        """
        Private method which performs the heavy lifting of parsing, called from
        ``parse()``, which passes on its ``kwargs`` to this function.

        :param timestr:
            The string to parse.

        :param dayfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the day (``True``) or month (``False``). If
            ``yearfirst`` is set to ``True``, this distinguishes between YDM
            and YMD. If set to ``None``, this value is retrieved from the
            current :class:`parserinfo` object (which itself defaults to
            ``False``).

        :param yearfirst:
            Whether to interpret the first value in an ambiguous 3-integer date
            (e.g. 01/05/09) as the year. If ``True``, the first number is taken
            to be the year, otherwise the last number is taken to be the year.
            If this is set to ``None``, the value is retrieved from the current
            :class:`parserinfo` object (which itself defaults to ``False``).

        :param fuzzy:
            Whether to allow fuzzy parsing, allowing for string like "Today is
            January 1, 2047 at 8:21:00AM".

        :param fuzzy_with_tokens:
            If ``True``, ``fuzzy`` is automatically set to True, and the parser
            will return a tuple where the first element is the parsed
            :class:`datetime.datetime` datetimestamp and the second element is
            a tuple containing the portions of the string which were ignored:

            .. doctest::

                >>> from dateutil.parser import parse
                >>> parse("Today is January 1, 2047 at 8:21:00AM", fuzzy_with_tokens=True)
                (datetime.datetime(2047, 1, 1, 8, 21), (u'Today is ', u' ', u'at '))

        """
        if fuzzy_with_tokens:
            fuzzy = True

        info = self.info

        if dayfirst is None:
            dayfirst = info.dayfirst

        if yearfirst is None:
            yearfirst = info.yearfirst

        res = self._result()
        l = _timelex.split(timestr)  # Splits the timestr into tokens

        skipped_idxs = []

        # year/month/day list
        ymd = _ymd()

        len_l = len(l)
        i = 0
        try:
            while i < len_l:

                # Check if it's a number
                value_repr = l[i]
                try:
                    value = float(value_repr)
                except ValueError:
                    value = None

                if value is not None:
                    # Numeric token
                    i = self._parse_numeric_token(l, i, info, ymd, res, fuzzy)

                # Check weekday
                elif info.weekday(l[i]) is not None:
                    value = info.weekday(l[i])
                    res.weekday = value

                # Check month name
                elif info.month(l[i]) is not None:
                    value = info.month(l[i])
                    ymd.append(value, 'M')

                    if i + 1 < len_l:
                        if l[i + 1] in ('-', '/'):
                            # Jan-01[-99]
                            sep = l[i + 1]
                            ymd.append(l[i + 2])

                            if i + 3 < len_l and l[i + 3] == sep:
                                # Jan-01-99
                                ymd.append(l[i + 4])
                                i += 2

                            i += 2

                        elif (i + 4 < len_l and l[i + 1] == l[i + 3] == ' ' and
                              info.pertain(l[i + 2])):
                            # Jan of 01
                            # In this case, 01 is clearly year
                            if l[i + 4].isdigit():
                                # Convert it here to become unambiguous
                                value = int(l[i + 4])
                                year = str(info.convertyear(value))
                                ymd.append(year, 'Y')
                            else:
                                # Wrong guess
                                pass
                                # TODO: not hit in tests
                            i += 4

                # Check am/pm
                elif info.ampm(l[i]) is not None:
                    value = info.ampm(l[i])
                    val_is_ampm = self._ampm_valid(res.hour, res.ampm, fuzzy)

                    if val_is_ampm:
                        res.hour = self._adjust_ampm(res.hour, value)
                        res.ampm = value

                    elif fuzzy:
                        skipped_idxs.append(i)

                # Check for a timezone name
                elif self._could_be_tzname(res.hour, res.tzname, res.tzoffset, l[i]):
                    res.tzname = l[i]
                    res.tzoffset = info.tzoffset(res.tzname)

                    # Check for something like GMT+3, or BRST+3. Notice
                    # that it doesn't mean "I am 3 hours after GMT", but
                    # "my time +3 is GMT". If found, we reverse the
                    # logic so that timezone parsing code will get it
                    # right.
                    if i + 1 < len_l and l[i + 1] in ('+', '-'):
                        l[i + 1] = ('+', '-')[l[i + 1] == '+']
                        res.tzoffset = None
                        if info.utczone(res.tzname):
                            # With something like GMT+3, the timezone
                            # is *not* GMT.
                            res.tzname = None

                # Check for a numbered timezone
                elif res.hour is not None and l[i] in ('+', '-'):
                    signal = (-1, 1)[l[i] == '+']
                    len_li = len(l[i + 1])

                    # TODO: check that l[i + 1] is integer?
                    if len_li == 4:
                        # -0300
                        hour_offset = int(l[i + 1][:2])
                        min_offset = int(l[i + 1][2:])
                    elif i + 2 < len_l and l[i + 2] == ':':
                        # -03:00
                        hour_offset = int(l[i + 1])
                        min_offset = int(l[i + 3])  # TODO: Check that l[i+3] is minute-like?
                        i += 2
                    elif len_li <= 2:
                        # -[0]3
                        hour_offset = int(l[i + 1][:2])
                        min_offset = 0
                    else:
                        raise ValueError(timestr)

                    res.tzoffset = signal * (hour_offset * 3600 + min_offset * 60)

                    # Look for a timezone name between parenthesis
                    if (i + 5 < len_l and
                            info.jump(l[i + 2]) and l[i + 3] == '(' and
                            l[i + 5] == ')' and
                            3 <= len(l[i + 4]) and
                            self._could_be_tzname(res.hour, res.tzname,
                                                  None, l[i + 4])):
                        # -0300 (BRST)
                        res.tzname = l[i + 4]
                        i += 4

                    i += 1

                # Check jumps
                elif not (info.jump(l[i]) or fuzzy):
                    raise ValueError(timestr)

                else:
                    skipped_idxs.append(i)
                i += 1

            # Process year/month/day
            year, month, day = ymd.resolve_ymd(yearfirst, dayfirst)

            res.century_specified = ymd.century_specified
            res.year = year
            res.month = month
            res.day = day
        ### CHANGES FROM HERE DOWN TO 3 VALUES IN TUPLE
        except (IndexError, ValueError):
            return None, None, None

        if not info.validate(res):
            return None, None, None

        if fuzzy_with_tokens:
            skipped_tokens, date_tokens = self._recombine_skipped_date(l, skipped_idxs)
            return res, tuple(skipped_tokens), tuple(date_tokens)
        else:
            return res, None, None

    def _recombine_skipped_date(self, tokens, skipped_idxs):
        """
        >>> tokens = ["foo", " ", "bar", " ", "19June2000", "baz"]
        >>> skipped_idxs = [0, 1, 2, 5]
        >>> _recombine_skipped_date(tokens, skipped_idxs)
        ["foo bar", "baz"], ["19June2000"]
        """
        skipped_tokens = []
        date_tokens = []
        prev = None
        for idx, token in enumerate(tokens):
            if idx in skipped_idxs:
                if prev is None or prev == 'date':
                    skipped_tokens.append(token)
                else:
                    skipped_tokens[-1] = skipped_tokens[-1] + token
                prev = 'skipped'
            else:
                if prev is None or prev == 'skipped':
                    date_tokens.append(token)
                else:
                    date_tokens[-1] = date_tokens[-1] + token
                prev = 'date'

        return skipped_tokens, date_tokens


DEFAULTPARSER = DateParser(parserinfo(dayfirst=True))


def parse(timestr, default=None,
          ignoretz=False, tzinfos=None, **kwargs):  # pragma: no cover
    """
    Parse the date/time string into a :class:`datetime.datetime` object.

    :param timestr:
        Any date/time string using the supported formats.

    :param default:
        The default datetime object, if this is a datetime object and not
        ``None``, elements specified in ``timestr`` replace elements in the
        default object.

    :param ignoretz:
        If set ``True``, time zones in parsed strings are ignored and a
        naive :class:`datetime.datetime` object is returned.

    :param tzinfos:
        Additional time zone names / aliases which may be present in the
        string. This argument maps time zone names (and optionally offsets
        from those time zones) to time zones. This parameter can be a
        dictionary with timezone aliases mapping time zone names to time
        zones or a function taking two parameters (``tzname`` and
        ``tzoffset``) and returning a time zone.

        The timezones to which the names are mapped can be an integer
        offset from UTC in seconds or a :class:`tzinfo` object.

        .. doctest::
           :options: +NORMALIZE_WHITESPACE

            >>> from dateutil.parser import parse
            >>> from dateutil.tz import gettz
            >>> tzinfos = {"BRST": -7200, "CST": gettz("America/Chicago")}
            >>> parse("2012-01-19 17:21:00 BRST", tzinfos=tzinfos)
            datetime.datetime(2012, 1, 19, 17, 21, tzinfo=tzoffset(u'BRST', -7200))
            >>> parse("2012-01-19 17:21:00 CST", tzinfos=tzinfos)
            datetime.datetime(2012, 1, 19, 17, 21,
                              tzinfo=tzfile('/usr/share/zoneinfo/America/Chicago'))

        This parameter is ignored if ``ignoretz`` is set.

    :param \\*\\*kwargs:
        Keyword arguments as passed to ``_parse()``.

    :return:
        Returns a :class:`datetime.datetime` object or, if the
        ``fuzzy_with_tokens`` option is ``True``, returns a tuple, the
        first element being a :class:`datetime.datetime` object, the second
        a tuple containing the fuzzy tokens.

    :raises ParserError:
        Raised for invalid or unknown string format, if the provided
        :class:`tzinfo` is not in a valid format, or if an invalid date
        would be created.

    :raises TypeError:
        Raised for non-string or character stream input.

    :raises OverflowError:
        Raised if the parsed date exceeds the largest valid C integer on
        your system.
    """

    if default is None:
        default = datetime.now().replace(hour=0, minute=0,
                                         second=0, microsecond=0)

    res, skipped_tokens, date_tokens = DEFAULTPARSER._parse(timestr, **kwargs)

    if res is None:
        raise ParserError("Unknown string format: %s", timestr)

    if len(res) == 0:
        raise ParserError("String does not contain a date: %s", timestr)

    try:
        ret = DEFAULTPARSER._build_naive(res, default)
    except ValueError as e:
        raisefrom(ParserError, str(e) + ": %s" % timestr, e)

    if not ignoretz:
        ret = DEFAULTPARSER._build_tzaware(ret, res, tzinfos)

    if kwargs.get('fuzzy_with_tokens', False):
        return ret, skipped_tokens, date_tokens
    else:
        return ret


def parse_date_range(string, date_format=None, fuzzy=None, zero_time=False):
    # type: (str, Optional[str], Optional[Dict], bool) -> Tuple[datetime,datetime]
    """Parse date from string using specified format (setting any time elements to 0 if zero_time is True).
    If no format is supplied, the function will guess. For unambiguous formats, this should be fine.
    Returns date range in dictionary keys startdate and enddate. If a dictionary is supplied in the fuzzy parameter,
    then dateutil's fuzzy parsing is used and the results returned in the dictionary in keys startdate, enddate,
    date (the string elements used to make the date) and nondate (the non date part of the string).

    Args:
        string (str): Dataset date string
        date_format (Optional[str]): Date format. If None is given, will attempt to guess. Defaults to None.
        fuzzy (Optional[Dict]): If dict supplied, fuzzy matching will be used and results returned in dict
        zero_time (bool): Zero time elements of datetime if True. Defaults to False.

    Returns:
        Tuple[datetime,datetime]: Tuple containing start date and end date
    """
    if date_format is None or fuzzy is not None:
        if fuzzy is not None:
            parsed_string1 = parse(string, fuzzy_with_tokens=True, default=default_date)
            parsed_string2 = parse(string, fuzzy_with_tokens=True, default=default_enddate)
            startdate = parsed_string1[0]
            enddate = parsed_string2[0]
            nondate = parsed_string1[1]
            if nondate:
                fuzzy['nondate'] = nondate
            else:
                fuzzy['nondate'] = None
            fuzzy['date'] = parsed_string1[2]
        else:
            startdate = parse(string, default=default_date)
            enddate = parse(string, default=default_enddate)
        if startdate.year == default_sd_year and enddate.year == default_ed_year:
            raise ParserError('No year in date!')
    else:
        try:
            startdate = datetime.strptime(string, date_format)
        except ValueError as e:
            raisefrom(ParserError, str(e), e)
        if startdate.year == 1900 and '%Y' not in date_format:  # 1900 is default when no year supplied
            raise ParserError('No year in date!')
        enddate = startdate
        if not any(str in date_format for str in ['%d', '%j']):
            startdate = startdate.replace(day=default_date.day)
            endday = default_enddate.day
            not_set = True
            while not_set:
                try:
                    enddate = enddate.replace(day=endday)
                    not_set = False
                except ValueError as e:
                    endday -= 1
                    if endday == 0:
                        raisefrom(ParserError, 'No end day of month found for %s!' % str(enddate), e)
        if not any(str in date_format for str in ['%b', '%B', '%m', '%j']):
            startdate = startdate.replace(month=default_date.month)
            enddate = enddate.replace(month=default_enddate.month)
    if zero_time:
        startdate = startdate.replace(hour=0, minute=0, second=0, microsecond=0)
        enddate = enddate.replace(hour=0, minute=0, second=0, microsecond=0)
    if fuzzy is not None:
        fuzzy['startdate'] = startdate
        fuzzy['enddate'] = enddate
    return startdate, enddate


def parse_date(string, date_format=None, fuzzy=None, zero_time=False):
    # type: (str, Optional[str], Optional[Dict], bool) -> datetime
    """Parse date from string using specified format (setting any time elements to 0 if zero_time is True).
    If no format is supplied, the function will guess. For unambiguous formats, this should be fine.
    Returns a datetime object. Raises exception for dates that are missing year, month or day.
    If a dictionary is supplied in the fuzzy parameter, then dateutil's fuzzy parsing is used and the results
    returned in the dictionary in keys startdate, enddate, date (the string elements used to make the date) and
    nondate (the non date part of the string).

    Args:
        string (str): Dataset date string
        date_format (Optional[str]): Date format. If None is given, will attempt to guess. Defaults to None.
        fuzzy (Optional[Dict]): If dict supplied, fuzzy matching will be used and results returned in dict
        zero_time (bool): Zero time elements of datetime if True. Defaults to False.

    Returns:
        datetime: The parsed date
    """
    startdate, enddate = parse_date_range(string, date_format=date_format, fuzzy=fuzzy, zero_time=zero_time)
    if startdate != enddate:
        raise ParserError('date is not a specific day!')
    return startdate



# -*- coding: utf-8 -*-
"""Saving utilities for YAML, JSON etc."""
import json
from io import open
from typing import Dict

import pyaml
import six
import yaml
import yamlloader


def save_str_to_file(string, path, encoding='utf-8'):
    # type: (str, str, str) -> None
    """Save string to file

    Args:
        string (str): String to save
        path (str): Path to file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        None
    """
    with open(path, 'w', encoding=encoding) as f:
        if six.PY2:
            string = unicode(string)
        f.write(string)


def save_yaml(dictionary, path, encoding='utf-8', pretty=False, sortkeys=False):
    # type: (Dict, str, str, bool, bool) -> None
    """Save dictionary to YAML file preserving order if it is an OrderedDict

    Args:
        dictionary (Dict): Python dictionary to save
        path (str): Path to YAML file
        encoding (str): Encoding of file. Defaults to utf-8.
        pretty (bool): Whether to pretty print. Defaults to False.
        sortkeys (bool): Whether to sort dictionary keys. Defaults to False.

    Returns:
        None
    """
    if sortkeys:
        dictionary = dict(dictionary)
    with open(path, 'w', encoding=encoding) as f:
        if pretty:
            pyaml.dump(dictionary, f)
        else:
            yaml.dump(dictionary, f, default_flow_style=None, Dumper=yamlloader.ordereddict.CDumper)


def save_json(dictionary, path, encoding='utf-8', pretty=False, sortkeys=False):
    # type: (Dict, str, str, bool, bool) -> None
    """Save dictionary to JSON file preserving order if it is an OrderedDict

    Args:
        dictionary (Dict): Python dictionary to save
        path (str): Path to JSON file
        encoding (str): Encoding of file. Defaults to utf-8.
        pretty (bool): Whether to pretty print. Defaults to False.
        sortkeys (bool): Whether to sort dictionary keys. Defaults to False.

    Returns:
        None
    """
    with open(path, 'w', encoding=encoding) as f:
        if pretty:
            indent = 2
            separators = (',', ': ')
        else:
            indent = None
            separators = (', ', ': ')
        if six.PY3:
            json.dump(dictionary, f, indent=indent, sort_keys=sortkeys, separators=separators)
        else:
            string = json.dumps(dictionary, ensure_ascii=False, indent=indent,
                                sort_keys=sortkeys, separators=separators)
            f.write(unicode(string))



# -*- coding: UTF-8 -*-
"""Loader Tests"""
from collections import OrderedDict
from os.path import join

import pytest

from hdx.utilities.loader import load_yaml, load_json, load_and_merge_yaml, load_and_merge_json, LoadError, \
    load_file_to_str, load_yaml_into_existing_dict, load_json_into_existing_dict
from hdx.utilities.path import temp_dir
from hdx.utilities.saver import save_str_to_file


class TestLoader:
    expected_yaml = OrderedDict([('hdx_prod_site', OrderedDict([('url', 'https://data.humdata.org/'),
                                                                ('username', None), ('password', None)])),
                                 ('hdx_test_site', OrderedDict([('url', 'https://test-data.humdata.org/'),
                                                                ('username', 'lala'), ('password', 'lalala')])),
                                 ('dataset', OrderedDict([('required_fields', ['name', 'title', 'dataset_date'])])),
                                 (
                                     'resource',
                                     OrderedDict([('required_fields', ['package_id', 'name', 'description'])])),
                                 ('showcase', OrderedDict([('required_fields', ['name', 'title'])])),
                                 ('param_1', 'ABC')])

    expected_json = OrderedDict([('hdx_prod_site', OrderedDict([('url', 'https://data.humdata.org/'),
                                                                ('username', None), ('password', None)])),
                                 ('hdx_test_site', OrderedDict([('url', 'https://test-data.humdata.org/'),
                                                                ('username', 'tumteetum'),
                                                                ('password', 'tumteetumteetum')])),
                                 ('dataset', OrderedDict([('required_fields', ['name', 'dataset_date'])])),
                                 ('resource', OrderedDict([('required_fields', ['name', 'description'])])),
                                 ('showcase', OrderedDict([('required_fields', ['name'])])), ('my_param', 'abc')])
    text = '''  hello
this
is
a
test  

'''
    expected_text_strip = '''hello
this
is
a
test'''
    expected_text_newlines_to_spaces = '''  hello this is a test    '''

    def test_load_empty(self, fixturesfolder):
        loaderfolder = join(fixturesfolder, 'loader')
        with pytest.raises(LoadError):
            load_file_to_str(join(loaderfolder, 'empty.yml'))
        with pytest.raises(LoadError):
            load_yaml(join(loaderfolder, 'empty.yml'))
        with pytest.raises(LoadError):
            load_json(join(loaderfolder, 'empty.json'))

    def test_load_and_merge_yaml(self, configfolder):
        result = load_and_merge_yaml([join(configfolder, 'hdx_config.yml'),
                                      join(configfolder, 'project_configuration.yml')])
        assert result == TestLoader.expected_yaml

    def test_load_and_merge_json(self, configfolder):
        result = load_and_merge_json([join(configfolder, 'hdx_config.json'),
                                      join(configfolder, 'project_configuration.json')])
        assert result == TestLoader.expected_json

    def test_load_yaml_into_existing_dict(self, configfolder):
        existing_dict = load_yaml(join(configfolder, 'hdx_config.yml'))
        result = load_yaml_into_existing_dict(existing_dict,
                                              join(configfolder, 'project_configuration.yml'))
        assert result == TestLoader.expected_yaml

    def test_load_json_into_existing_dict(self, configfolder):
        existing_dict = load_json(join(configfolder, 'hdx_config.json'))
        result = load_json_into_existing_dict(existing_dict,
                                              join(configfolder, 'project_configuration.json'))
        assert result == TestLoader.expected_json

    def test_load_file_to_str(self):
        with temp_dir(folder='test_text') as tmpdir:
            text_file = join(tmpdir, 'text_file.txt')
            save_str_to_file(TestLoader.text, text_file)
            result = load_file_to_str(text_file)
            assert result == TestLoader.text
            result = load_file_to_str(text_file, strip=True)
            assert result == TestLoader.expected_text_strip
            result = load_file_to_str(text_file, replace_newlines=' ')
            assert result == TestLoader.expected_text_newlines_to_spaces
            with pytest.raises(IOError):
                load_file_to_str(join(tmpdir, 'NOTEXIST.txt'))



# -*- coding: utf-8 -*-
"""User agent utilities"""
import logging
import os
from os.path import join, expanduser, isfile
from typing import Any, Dict, Optional

from hdx.utilities.loader import load_yaml
from hdx.utilities.version import get_utils_version

logger = logging.getLogger(__name__)


class UserAgentError(Exception):
    pass


class UserAgent(object):
    default_user_agent_config_yaml = join(expanduser('~'), '.useragent.yml')
    user_agent = None

    @staticmethod
    def _environment_variables(**kwargs):
        # type: (Any) -> Any
        """
        Overwrite keyword arguments with environment variables

        Args:
            **kwargs: See below
            user_agent (str): User agent string.

        Returns:
            kwargs: Changed keyword arguments

        """
        user_agent = os.getenv('USER_AGENT')
        if user_agent is not None:
            kwargs['user_agent'] = user_agent
        preprefix = os.getenv('PREPREFIX')
        if preprefix is not None:
            kwargs['preprefix'] = preprefix
        return kwargs

    @staticmethod
    def _construct(configdict, prefix, ua):
        # type: (Dict, str, str) -> str
        """
        Construct user agent

        Args:
            configdict (str): Additional configuration for user agent
            prefix (str): Text to put at start of user agent
            ua (str): Custom user agent text

        Returns:
            str: Full user agent string

        """
        if not ua:
            raise UserAgentError("User_agent parameter missing. It can be your project's name for example.")
        preprefix = configdict.get('preprefix')
        if preprefix:
            user_agent = '%s:' % preprefix
        else:
            user_agent = ''
        if prefix:
            user_agent = '%s%s-' % (user_agent, prefix)
        user_agent = '%s%s' % (user_agent, ua)
        return user_agent

    @classmethod
    def _load(cls, prefix, user_agent_config_yaml, user_agent_lookup=None):
        # type: (str, str, Optional[str]) -> str
        """
        Load user agent YAML file

        Args:
            prefix (str): Text to put at start of user agent
            user_agent_config_yaml (str): Path to user agent YAML file
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            str: user agent

        """
        if not user_agent_config_yaml:
            user_agent_config_yaml = cls.default_user_agent_config_yaml
            logger.info(
                'No user agent or user agent config file given. Using default user agent config file: %s.' % user_agent_config_yaml)
        if not isfile(user_agent_config_yaml):
            raise UserAgentError(
                "User_agent should be supplied in a YAML config file. It can be your project's name for example.")
        logger.info('Loading user agent config from: %s' % user_agent_config_yaml)
        user_agent_config_dict = load_yaml(user_agent_config_yaml)
        if user_agent_lookup:
            user_agent_config_dict = user_agent_config_dict.get(user_agent_lookup)
        if not user_agent_config_dict:
            raise UserAgentError("No user agent information read from: %s" % user_agent_config_yaml)
        ua = user_agent_config_dict.get('user_agent')
        return cls._construct(user_agent_config_dict, prefix, ua)

    @classmethod
    def _create(cls, user_agent=None, user_agent_config_yaml=None,
                user_agent_lookup=None, **kwargs):
        # type: (Optional[str], Optional[str], Optional[str], Any) -> str
        """
        Get full user agent string

        Args:
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            str: Full user agent string

        """
        kwargs = UserAgent._environment_variables(**kwargs)
        if 'user_agent' in kwargs:
            user_agent = kwargs['user_agent']
            del kwargs['user_agent']
        prefix = kwargs.get('prefix')
        if prefix:
            del kwargs['prefix']
        else:
            prefix = 'HDXPythonUtilities/%s' % get_utils_version()
        if not user_agent:
            ua = cls._load(prefix, user_agent_config_yaml, user_agent_lookup)
        else:
            ua = cls._construct(kwargs, prefix, user_agent)
        return ua

    @classmethod
    def clear_global(cls):
        # type: () -> None
        """
        Clear stored user agent string

        Returns:
            None

        """
        cls.user_agent = None

    @classmethod
    def set_global(cls, user_agent=None, user_agent_config_yaml=None,
                   user_agent_lookup=None, **kwargs):
        # type: (Optional[str], Optional[str], Optional[str], Any) -> None
        """
        Set global user agent string

        Args:
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            None
        """
        cls.user_agent = cls._create(user_agent, user_agent_config_yaml, user_agent_lookup, **kwargs)

    @classmethod
    def get(cls, user_agent=None, user_agent_config_yaml=None, user_agent_lookup=None, **kwargs):
        # type: (Optional[str], Optional[str], Optional[str], Any) -> str
        """
        Get full user agent string from parameters if supplied falling back on global user agent if set.

        Args:
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            str: Full user agent string

        """
        if user_agent or user_agent_config_yaml or 'user_agent' in UserAgent._environment_variables(**kwargs):
            return UserAgent._create(user_agent, user_agent_config_yaml, user_agent_lookup, **kwargs)
        if cls.user_agent:
            return cls.user_agent
        else:
            raise UserAgentError(
                'You must either set the global user agent: UserAgent.set_global(...) or pass in user agent parameters!')



# -*- coding: UTF-8 -*-
"""Downloader Tests"""
import copy
import re
from collections import OrderedDict
from os import remove
from os.path import join, abspath
from shutil import rmtree, copytree
from tempfile import gettempdir

import pytest

from hdx.utilities.downloader import Download, DownloadError
from hdx.utilities.session import SessionError
from hdx.utilities.useragent import UserAgent


class TestDownloader:
    downloaderfoldername = 'downloader'

    @pytest.fixture(scope='class', autouse=True)
    def useragent(self):
        UserAgent.set_global('test')
        yield
        UserAgent.clear_global()

    @pytest.fixture(scope='class')
    def downloaderfolder(self, fixturesfolder):
        return join(fixturesfolder, self.downloaderfoldername)

    @pytest.fixture(scope='class')
    def fixtureurl(self):
        return 'https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/master/tests/fixtures/test_data.csv'

    @pytest.fixture(scope='class')
    def getfixtureurl(self):
        return 'http://httpbin.org/get'

    @pytest.fixture(scope='class')
    def postfixtureurl(self):
        return 'http://httpbin.org/post'

    @pytest.fixture(scope='class')
    def fixturenotexistsurl(self):
        return 'https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/master/tests/fixtures/NOTEXIST.csv'

    @pytest.fixture(scope='class')
    def fixtureprocessurl(self):
        return 'https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/master/tests/fixtures/downloader/test_csv_processing.csv?a=1'

    @pytest.fixture(scope='class')
    def fixtureprocessurlblank(self):
        return 'https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/master/tests/fixtures/downloader/test_csv_processing_blanks.csv?a=1'

    def test_get_path_for_url(self, tmpdir, fixtureurl, configfolder, downloaderfolder):
        tmpdir = str(tmpdir)
        filename = 'test_data.csv'
        path = Download.get_path_for_url(fixtureurl, configfolder)
        assert abspath(path) == abspath(join(configfolder, filename))
        path = Download.get_path_for_url(fixtureurl, downloaderfolder)
        assert abspath(path) == abspath(join(downloaderfolder, 'test_data3.csv'))
        testfolder = join(tmpdir, self.downloaderfoldername)
        rmtree(testfolder, ignore_errors=True)
        copytree(downloaderfolder, testfolder)
        path = Download.get_path_for_url(fixtureurl, testfolder, overwrite=True)
        assert abspath(path) == abspath(join(testfolder, filename))
        rmtree(testfolder)
        filename = 'myfilename.txt'
        path = Download.get_path_for_url(fixtureurl, filename=filename)
        assert abspath(path) == abspath(join(gettempdir(), filename))
        path = Download.get_path_for_url(fixtureurl, downloaderfolder, filename)
        assert abspath(path) == abspath(join(downloaderfolder, filename))

    def test_init(self, monkeypatch, downloaderfolder):
        basicauthfile = join(downloaderfolder, 'basicauth.txt')
        with Download(auth=('u', 'p')) as downloader:
            assert downloader.session.auth == ('u', 'p')
        basicauth = 'Basic dXNlcjpwYXNz'
        with Download(basic_auth=basicauth) as downloader:
            assert downloader.session.auth == ('user', 'pass')
        with Download(basic_auth_file=basicauthfile) as downloader:
            assert downloader.session.auth == ('testuser', 'testpass')
        extraparamsyamltree = join(downloaderfolder, 'extra_params_tree.yml')
        with Download(basic_auth='Basic dXNlcjpwYXNz', extra_params_yaml=extraparamsyamltree,
                      extra_params_lookup='mykey') as downloader:
            assert downloader.session.auth == ('user', 'pass')
        monkeypatch.setenv('BASIC_AUTH', basicauth)
        with Download(basic_auth='12345') as downloader:
            assert downloader.session.auth == ('user', 'pass')
        with Download(extra_params_yaml=extraparamsyamltree, extra_params_lookup='mykey') as downloader:
            assert downloader.session.auth == ('user', 'pass')
        monkeypatch.delenv('BASIC_AUTH')
        with pytest.raises(SessionError):
            Download(auth=('u', 'p'), basic_auth='Basic xxxxxxxxxxxxxxxx')
        with pytest.raises(SessionError):
            Download(auth=('u', 'p'), basic_auth_file=join('lala', 'lala.txt'))
        with pytest.raises(SessionError):
            Download(basic_auth='Basic dXNlcjpwYXNz', basic_auth_file=join('lala', 'lala.txt'))
        with pytest.raises(SessionError):
            Download(auth=('u', 'p'), extra_params_yaml=extraparamsyamltree, extra_params_lookup='mykey')
        with pytest.raises(SessionError):
            Download(basic_auth_file=basicauthfile, extra_params_yaml=extraparamsyamltree, extra_params_lookup='mykey')
        with pytest.raises(SessionError):
            Download(extra_params_yaml=extraparamsyamltree, extra_params_lookup='missingkey')
        with pytest.raises(IOError):
            Download(basic_auth_file='NOTEXIST')
        extraparamsjson = join(downloaderfolder, 'extra_params.json')
        extraparamsyaml = join(downloaderfolder, 'extra_params.yml')
        test_url = 'http://www.lalala.com/lala'
        with Download(basic_auth_file=basicauthfile, extra_params_dict={'key1': 'val1'}) as downloader:
            assert downloader.session.auth == ('testuser', 'testpass')
            assert downloader.get_full_url(test_url) == '%s?key1=val1' % test_url
        with Download(extra_params_json=extraparamsjson) as downloader:
            full_url = downloader.get_full_url(test_url)
            assert 'param_1=value+1' in full_url
            assert 'param_2=value_2' in full_url
            assert 'param_3=12' in full_url
        with Download(extra_params_yaml=extraparamsyaml) as downloader:
            full_url = downloader.get_full_url(test_url)
            assert 'param1=value1' in full_url
            assert 'param2=value+2' in full_url
            assert 'param3=10' in full_url
        with Download(extra_params_yaml=extraparamsyamltree, extra_params_lookup='mykey') as downloader:
            assert downloader.session.auth == ('testuser', 'testpass')
            full_url = downloader.get_full_url(test_url)
            assert 'param1=value+1' in full_url
            assert 'param2=value2' in full_url
            assert 'param3=11' in full_url
            assert 'basic_auth' not in full_url
        monkeypatch.setenv('EXTRA_PARAMS', 'param1=value2+3,param2=value3+7')
        with Download(extra_params_yaml=extraparamsyamltree, extra_params_lookup='mykey') as downloader:
            assert downloader.session.auth is None
            full_url = downloader.get_full_url(test_url)
            assert 'param1=value2%2B3' in full_url
            assert 'param2=value3%2B7' in full_url
            assert 'param3=11' not in full_url
            assert 'basic_auth' not in full_url
        with Download(extra_params_yaml=extraparamsyamltree, extra_params_lookup='mykey', use_env=False) as downloader:
            assert downloader.session.auth == ('testuser', 'testpass')
            full_url = downloader.get_full_url(test_url)
            assert 'param1=value+1' in full_url
            assert 'param2=value2' in full_url
            assert 'param3=11' in full_url
            assert 'basic_auth' not in full_url
        monkeypatch.delenv('EXTRA_PARAMS')
        with pytest.raises(SessionError):
            Download(extra_params_dict={'key1': 'val1'}, extra_params_json=extraparamsjson)
        with pytest.raises(SessionError):
            Download(extra_params_dict={'key1': 'val1'}, extra_params_yaml=extraparamsyaml)
        with pytest.raises(SessionError):
            Download(extra_params_dict={'key1': 'val1'}, extra_params_yaml=extraparamsyamltree,
                     extra_params_lookup='mykey')
        with pytest.raises(IOError):
            Download(extra_params_json='NOTEXIST')

    def test_get_url_for_get(self):
        assert Download.get_url_for_get('http://www.lala.com/hdfa?a=3&b=4', OrderedDict(
            [('c', 'e'), ('d', 'f')])) == 'http://www.lala.com/hdfa?a=3&b=4&c=e&d=f'

    def test_get_url_params_for_post(self):
        assert Download.get_url_params_for_post('http://www.lala.com/hdfa?a=3&b=4',
                                                OrderedDict([('c', 'e'), ('d', 'f')])) == (
                   'http://www.lala.com/hdfa', OrderedDict([('a', '3'), ('b', '4'), ('c', 'e'), ('d', 'f')]))

    def test_hxl_row(self):
        headers = ['a', 'b', 'c']
        hxltags = {'b': '#b', 'c': '#c'}
        assert Download.hxl_row(headers, hxltags) == ['', '#b', '#c']
        assert Download.hxl_row(headers, hxltags, dict_form=True) == {'a': '', 'b': '#b', 'c': '#c'}
        assert Download.hxl_row(headers, dict()) == ['', '', '']
        assert Download.hxl_row([], hxltags) == list()

    def test_setup_stream(self, fixtureurl, fixturenotexistsurl, getfixtureurl, postfixtureurl):
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.setup('NOTEXIST://NOTEXIST.csv')
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.setup(fixturenotexistsurl)
        with Download() as downloader:
            downloader.setup(fixtureurl)
            headers = downloader.response.headers
            assert headers['Content-Length'] == '742'
        with Download() as downloader:
            downloader.setup(postfixtureurl, post=True)
            headers = downloader.response.headers
            assert bool(re.match(r'4\d\d', headers['Content-Length'])) is True
            downloader.setup('%s?id=10&lala=a' % getfixtureurl, post=False,
                             parameters=OrderedDict([('b', '4'), ('d', '3')]))
            assert downloader.get_json()['args'] == OrderedDict([('b', '4'), ('d', '3'), ('id', '10'), ('lala', 'a')])
            downloader.setup('%s?id=3&lala=b' % postfixtureurl, post=True,
                             parameters=OrderedDict([('a', '3'), ('c', '2')]))
            assert downloader.get_json()['form'] == OrderedDict([('a', '3'), ('c', '2'), ('id', '3'), ('lala', 'b')])

    def test_hash_stream(self, fixtureurl):
        with Download() as downloader:
            downloader.setup(fixtureurl)
            md5hash = downloader.hash_stream(fixtureurl)
            assert md5hash == 'da9db35a396cca10c618f6795bdb9ff2'

    def test_download_file(self, tmpdir, fixtureurl, fixturenotexistsurl, getfixtureurl, postfixtureurl):
        tmpdir = str(tmpdir)
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file('NOTEXIST://NOTEXIST.csv', tmpdir)
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(fixturenotexistsurl)
        with Download() as downloader:
            f = downloader.download_file(fixtureurl, folder=tmpdir)
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, 'test_data.csv'))
            filename = 'myfilename.txt'
            f = downloader.download_file(fixtureurl, folder=tmpdir, filename=filename)
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file('%s?id=10&lala=a' % getfixtureurl, post=False,
                                         parameters=OrderedDict([('b', '4'), ('d', '3')]), folder=tmpdir,
                                         filename=filename)
            fpath = abspath(f)
            with open(fpath, 'rt') as fi:
                text = fi.read()
                assert '"id": "10"' in text
                assert '"lala": "a"' in text
                assert '"b": "4"' in text
                assert '"d": "3"' in text
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file('%s?id=3&lala=b' % postfixtureurl, post=True,
                                         parameters=OrderedDict([('a', '3'), ('c', '2')]), folder=tmpdir,
                                         filename=filename)
            fpath = abspath(f)
            with open(fpath, 'rt') as fi:
                text = fi.read()
                assert '"id": "3"' in text
                assert '"lala": "b"' in text
                assert '"a": "3"' in text
                assert '"c": "2"' in text
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))

    def test_download(self, fixtureurl, fixturenotexistsurl, getfixtureurl, postfixtureurl):
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download('NOTEXIST://NOTEXIST.csv')
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download(fixturenotexistsurl)
        with Download() as downloader:
            result = downloader.download(fixtureurl)
            assert result.headers['Content-Length'] == '742'
            downloader.download('%s?id=10&lala=a' % getfixtureurl, post=False,
                                parameters=OrderedDict([('b', '4'), ('d', '3')]))
            assert downloader.get_json()['args'] == OrderedDict([('b', '4'), ('d', '3'), ('id', '10'), ('lala', 'a')])
            downloader.download('%s?id=3&lala=b' % postfixtureurl, post=True,
                                parameters=OrderedDict([('a', '3'), ('c', '2')]))
            assert downloader.get_json()['form'] == OrderedDict([('a', '3'), ('c', '2'), ('id', '3'), ('lala', 'b')])
        with Download(rate_limit={'calls': 1, 'period': 0.1}) as downloader:
            downloader.download('%s?id=10&lala=a' % getfixtureurl, post=False,
                                parameters=OrderedDict([('b', '4'), ('d', '3')]))
            assert downloader.get_json()['args'] == OrderedDict([('b', '4'), ('d', '3'), ('id', '10'), ('lala', 'a')])

    def test_download_tabular_key_value(self, fixtureurl, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_key_value(fixtureurl, file_type='csv')
            assert result == {'615': '2231RTA', 'GWNO': 'EVENT_ID_CNTY'}
            result = downloader.download_tabular_key_value(fixtureprocessurl, headers=2)
            assert result == {'coal': '3', 'gas': '2'}
            with pytest.raises(DownloadError):
                downloader.download_tabular_key_value('NOTEXIST://NOTEXIST.csv')

    @staticmethod
    def fix_strings(result):  # This isn't needed when I run locally but is when run in Travis!
        for x in result:
            for y in result[x]:
                result[x][y] = result[x][y].replace("'", '')

    def test_get_tabular_rows_as_list(self, fixtureprocessurl):
        with Download() as downloader:
            rows = list(downloader.get_tabular_rows_as_list(fixtureprocessurl))
            assert rows == [['la1', 'ha1', 'ba1', 'ma1'], ['header1', 'header2', 'header3', 'header4'],
                            ['coal', '3', '7.4', 'needed'], ['gas', '2', '6.5', 'n/a']]

    def test_get_tabular_rows(self, fixtureprocessurl, fixtureprocessurlblank):
        with Download() as downloader:
            expected = [['la1', 'ha1', 'ba1', 'ma1'], ['header1', 'header2', 'header3', 'header4'],
                        ['coal', '3', '7.4', 'needed'], ['gas', '2', '6.5', 'n/a']]
            expected_headers = expected[0]
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl)
            assert headers == expected_headers
            assert list(iterator) == expected[1:]
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurlblank)
            assert headers == expected_headers
            blank_expected = copy.deepcopy(expected[1:])
            blank_expected[2][0] = ''
            assert list(iterator) == blank_expected
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurlblank, ignore_blank_rows=False)
            assert headers == expected_headers
            blank_expected.append(list())
            blank_expected.insert(1, list())
            assert list(iterator) == blank_expected
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=1)
            assert headers == expected_headers
            assert list(iterator) == expected[1:]
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=2)
            assert headers == expected[1]
            assert list(iterator) == expected[2:]
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=[1, 2])
            assert headers == ['la1 header1', 'ha1 header2', 'ba1 header3', 'ma1 header4']
            assert list(iterator) == expected[2:]
            myheaders = ['a', 'b', 'c', 'd']
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=myheaders)
            assert headers == myheaders
            assert list(iterator) == expected
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=1, dict_form=True)
            assert headers == expected_headers
            expected_dicts = [{'la1': 'header1', 'ha1': 'header2', 'ba1': 'header3', 'ma1': 'header4'},
                              {'la1': 'coal', 'ha1': '3', 'ba1': '7.4', 'ma1': 'needed'},
                              {'la1': 'gas', 'ha1': '2', 'ba1': '6.5', 'ma1': 'n/a'}]
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=3, dict_form=True)
            assert headers == expected[2]
            expected_dicts = [{'coal': 'gas', '3': '2', '7.4': '6.5', 'needed': 'n/a'}]
            assert list(iterator) == expected_dicts

            def testfn(headers, row):
                row.insert(2, 'lala')
                return row

            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=3,
                                                            header_insertions=[(2, 'la')], row_function=testfn)
            expected_headers = ['coal', '3', 'la', '7.4', 'needed']
            assert headers == expected_headers
            assert list(iterator) == [['gas', '2', 'lala', '6.5', 'n/a']]

            def testfn(headers, row):
                row['la'] = 'lala'
                return row

            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl, headers=3, dict_form=True,
                                                            header_insertions=[(2, 'la')], row_function=testfn)
            assert headers == expected_headers
            expected_dicts[0]['la'] = 'lala'
            assert list(iterator) == expected_dicts

            headers, iterator = downloader.get_tabular_rows(fixtureprocessurlblank, headers=1, dict_form=True, ignore_blank_rows=False,
                                                            header_insertions=[(2, 'la')], row_function=testfn)
            expected_headers = ['la1', 'ha1', 'la', 'ba1', 'ma1']
            assert headers == expected_headers

            expected_dicts = [{'la1': 'header1', 'ha1': 'header2', 'ba1': 'header3', 'ma1': 'header4', 'la': 'lala'},
                              {'la': 'lala'},
                              {'la1': 'coal', 'ha1': '3', 'ba1': '7.4', 'ma1': 'needed', 'la': 'lala'},
                              {'la1': '', 'ha1': '2', 'ba1': '6.5', 'ma1': 'n/a', 'la': 'lala'},
                              {'la': 'lala'}]
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurlblank, headers=1, dict_form=True,
                                                            ignore_blank_rows=False, header_insertions=[(2, 'la')],
                                                            row_function=testfn)
            assert headers == expected_headers
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurlblank, headers=1, dict_form=True,
                                                            header_insertions=[(2, 'la')], row_function=testfn)
            assert headers == expected_headers
            del expected_dicts[4]
            del expected_dicts[1]
            assert list(iterator) == expected_dicts

            def testfn(headers, row):
                row['la'] = 'lala'
                if row['ha1'] == '3':
                    return None
                return row

            headers, iterator = downloader.get_tabular_rows(fixtureprocessurlblank, headers=1, dict_form=True,
                                                            header_insertions=[(2, 'la')], row_function=testfn)
            assert headers == expected_headers
            del expected_dicts[1]
            assert list(iterator) == expected_dicts

            with pytest.raises(DownloadError):
                downloader.get_tabular_rows(fixtureprocessurl, headers=None)

    def test_download_tabular_rows_as_dicts(self, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_rows_as_dicts(fixtureprocessurl, headers=2)
            self.fix_strings(result)
            assert result == {'coal': {'header2': '3', 'header3': '7.4', 'header4': 'needed'},
                              'gas': {'header2': '2', 'header3': '6.5', 'header4': 'n/a'}}
            result = downloader.download_tabular_rows_as_dicts(fixtureprocessurl, headers=2, keycolumn=2)
            self.fix_strings(result)
            assert result == {'2': {'header1': 'gas', 'header3': '6.5', 'header4': 'n/a'},
                              '3': {'header1': 'coal', 'header3': '7.4', 'header4': 'needed'}}

    def test_download_tabular_cols_as_dicts(self, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_cols_as_dicts(fixtureprocessurl, headers=2)
            self.fix_strings(result)
            assert result == {'header2': {'coal': '3', 'gas': '2'},
                              'header3': {'coal': '7.4', 'gas': '6.5'},
                              'header4': {'coal': 'needed', 'gas': 'n/a'}}
            result = downloader.download_tabular_cols_as_dicts(fixtureprocessurl, headers=2, keycolumn=2)
            self.fix_strings(result)
            assert result == {'header1': {'3': 'coal', '2': 'gas'},
                              'header3': {'3': '7.4', '2': '6.5'},
                              'header4': {'3': 'needed', '2': 'n/a'}}

    def test_get_column_positions(self):
        assert Download.get_column_positions(['a', 'b', 'c']) == {'a': 0, 'b': 1, 'c': 2}



# -*- coding: UTF-8 -*-
"""Database Utility Tests"""
import copy
import os
from collections import namedtuple
from os.path import join

import psycopg2
import pytest
from sshtunnel import SSHTunnelForwarder

from hdx.utilities.database import Database


class TestDatabase:
    connected = False
    started = False
    stopped = False
    dbpath = join('tests', 'test_database.db')
    params = {'database': 'mydatabase', 'host': 'myserver', 'port': 1234, 'username': 'myuser', 'password': 'mypass',
              'driver': 'postgres'}
    sqlalchemy_url = 'postgres://myuser:mypass@myserver:1234/mydatabase'

    @pytest.fixture(scope='function')
    def nodatabase(self):
        try:
            os.remove(TestDatabase.dbpath)
        except OSError:
            pass
        return 'sqlite:///%s' % TestDatabase.dbpath

    @pytest.fixture(scope='function')
    def mock_psycopg2(self, monkeypatch):
        def connect(**kwargs):
            if TestDatabase.connected:
                class Connection:
                    @staticmethod
                    def close():
                        return

                return Connection()
            else:
                TestDatabase.connected = True
                raise psycopg2.OperationalError

        monkeypatch.setattr(psycopg2, 'connect', connect)

    @pytest.fixture(scope='function')
    def mock_SSHTunnelForwarder(self, monkeypatch):
        def init(*args, **kwargs):
            return None

        def start(_):
            TestDatabase.started = True

        def stop(_):
            TestDatabase.stopped = True

        monkeypatch.setattr(SSHTunnelForwarder, '__init__', init)
        monkeypatch.setattr(SSHTunnelForwarder, 'start', start)
        monkeypatch.setattr(SSHTunnelForwarder, 'stop', stop)
        monkeypatch.setattr(SSHTunnelForwarder, 'local_bind_host', '0.0.0.0')
        monkeypatch.setattr(SSHTunnelForwarder, 'local_bind_port', 12345)

        def get_session(_, db_url):
            class Session:
                bind = namedtuple('Bind', 'engine')

                def close(self):
                    return None

            Session.bind.engine = namedtuple('Engine', 'url')
            Session.bind.engine.url = db_url
            return Session()

        monkeypatch.setattr(Database, 'get_session', get_session)

    def test_get_params_from_sqlalchemy_url(self):
        result = Database.get_params_from_sqlalchemy_url(TestDatabase.sqlalchemy_url)
        assert result == TestDatabase.params

    def test_get_sqlalchemy_url(self):
        result = Database.get_sqlalchemy_url(**TestDatabase.params)
        assert result == TestDatabase.sqlalchemy_url

    def test_wait_for_postgres(self, mock_psycopg2):
        TestDatabase.connected = False
        Database.wait_for_postgres('mydatabase', 'myserver', 5432, 'myuser', 'mypass')
        assert TestDatabase.connected is True

    def test_get_session(self, nodatabase):
        with Database(database=TestDatabase.dbpath, port=None, driver='sqlite') as dbsession:
            assert str(dbsession.bind.engine.url) == nodatabase

    def test_get_session_ssh(self, mock_psycopg2, mock_SSHTunnelForwarder):
        with Database(ssh_host='mysshhost', **TestDatabase.params) as dbsession:
            assert str(dbsession.bind.engine.url) == 'postgres://myuser:mypass@0.0.0.0:12345/mydatabase'
        params = copy.deepcopy(TestDatabase.params)
        del params['password']
        with Database(ssh_host='mysshhost', ssh_port=25, **params) as dbsession:
            assert str(dbsession.bind.engine.url) == 'postgres://myuser@0.0.0.0:12345/mydatabase'



# -*- coding: UTF-8 -*-
"""Compare Utility Tests"""
from os.path import join

import pytest

from hdx.utilities.compare import compare_files, assert_files_same


class TestCompare:
    @pytest.fixture(scope='class')
    def testfile1(self, fixturesfolder):
        return join(fixturesfolder, 'compare', 'test_csv_processing.csv')

    @pytest.fixture(scope='class')
    def testfile2(self, fixturesfolder):
        return join(fixturesfolder, 'compare', 'test_csv_processing2.csv')

    def test_compare_files(self, testfile1, testfile2):
        result = compare_files(testfile1, testfile2)
        assert result == ["- coal   ,3      ,7.4    ,'needed'", '?         ^\n',
                          "+ coal   ,1      ,7.4    ,'notneeded'", '?         ^                +++\n']
        assert_files_same(testfile1, testfile1)



# -*- coding: UTF-8 -*-
"""HTML Tests"""
from os.path import join

import pytest

from hdx.utilities.html import get_soup, extract_table
from hdx.utilities.loader import load_file_to_str


class TestHTML:
    url = 'https://unstats.un.org/unsd/methodology/m49/overview/'

    @pytest.fixture(scope='function')
    def htmltext(self, fixturesfolder):
        return load_file_to_str(join(fixturesfolder, 'html', 'response.html'))

    @pytest.fixture(scope='function')
    def downloader(self, htmltext):
        class Response:
            pass

        class Download:
            @staticmethod
            def download(url):
                response = Response()
                if url == TestHTML.url:
                    response.text = htmltext
                return response
        return Download()

    def test_html(self, downloader):
        soup = get_soup(TestHTML.url, downloader=downloader)
        tabletag = soup.find(id='downloadTableEN')
        result = extract_table(tabletag)
        assert len(result) == 249
        assert result[10] == {'Intermediate Region Name': 'Eastern Africa', 'Global Code': '001', 'Region Code': '002',
                              'Land Locked Developing Countries (LLDC)': '', 'Intermediate Region Code': '014',
                              'Sub-region Name': 'Sub-Saharan Africa', 'Least Developed Countries (LDC)': 'x',
                              'Global Name': 'World', 'M49 Code': '262', 'ISO-alpha3 Code': 'DJI',
                              'Small Island Developing States (SIDS)': '',
                              'Developed / Developing Countries': 'Developing', 'Country or Area': 'Djibouti',
                              'Region Name': 'Africa', 'Sub-region Code': '202'}
        assert result[247] == {'Global Code': '001', 'Sub-region Name': 'Polynesia', 'Intermediate Region Code': '',
                               'Intermediate Region Name': '', 'Region Code': '009',
                               'Least Developed Countries (LDC)': 'x', 'ISO-alpha3 Code': 'TUV',
                               'Sub-region Code': '061', 'Region Name': 'Oceania',
                               'Developed / Developing Countries': 'Developing',
                               'Small Island Developing States (SIDS)': 'x',
                               'Land Locked Developing Countries (LLDC)': '',
                               'M49 Code': '798', 'Global Name': 'World', 'Country or Area': 'Tuvalu'}



# -*- coding: UTF-8 -*-
"""Date Parse Utility Tests"""
from datetime import datetime
from parser import ParserError

import pytest

from hdx.utilities.dateparse import parse_date_range, parse_date


class TestDateParse:
    def test_parse_date_range(self):
        result = datetime(2013, 2, 10, 0, 0), datetime(2013, 2, 10, 0, 0)
        assert parse_date_range('10/02/2013') == result
        assert parse_date_range('2013/02/10') == result
        result = datetime(2013, 2, 20, 10, 0), datetime(2013, 2, 20, 10, 0)
        assert parse_date_range('20/02/2013 10:00:00') == result
        assert parse_date_range('20/02/2013 10:00:00', '%d/%m/%Y %H:%M:%S') == result
        result = datetime(2013, 2, 20, 0, 0), datetime(2013, 2, 20, 0, 0)
        assert parse_date_range('20/02/2013') == result
        assert parse_date_range('20/02/2013 10:00:00', zero_time=True) == result
        assert parse_date_range('20/02/2013', '%d/%m/%Y') == result
        assert parse_date_range('20/02/2013 10:00:00', '%d/%m/%Y %H:%M:%S', zero_time=True) == result
        fuzzy = dict()
        assert parse_date_range('date is 20/02/2013 for this test', fuzzy=fuzzy) == result
        fuzzyresult = {'startdate': datetime(2013, 2, 20, 0, 0), 'enddate': datetime(2013, 2, 20, 0, 0),
                       'nondate': ('date is ', ' for this test'), 'date': ('20/02/2013',)}
        assert fuzzy == fuzzyresult
        fuzzy = dict()
        assert parse_date_range('date is 20/02/2013 for this test', date_format='%d/%m/%Y', fuzzy=fuzzy) == result
        assert fuzzy == fuzzyresult
        fuzzy = dict()
        assert parse_date_range('20/02/2013', date_format='%d/%m/%Y', fuzzy=fuzzy) == result
        fuzzyresult['nondate'] = None
        assert fuzzy == fuzzyresult
        result = datetime(2013, 2, 1, 0, 0), datetime(2013, 2, 28, 0, 0)
        assert parse_date_range('02/2013') == result
        assert parse_date_range('02/2013', '%m/%Y') == result
        fuzzy = dict()
        assert parse_date_range('date is 02/2013 for this test', fuzzy=fuzzy) == result
        assert fuzzy == {'startdate': datetime(2013, 2, 1, 0, 0), 'enddate': datetime(2013, 2, 28, 0, 0),
                         'nondate': ('date is ', ' for this test'), 'date': ('02/2013',)}
        result = datetime(2013, 1, 1, 0, 0), datetime(2013, 12, 31, 0, 0)
        assert parse_date_range('2013') == result
        assert parse_date_range('2013', '%Y') == result
        fuzzy = dict()
        date = datetime(2001, 12, 10, 0, 0)
        result = date, date
        assert parse_date_range('State_Village_Tract_Boundaries 10/12/01 lala', fuzzy=fuzzy) == result
        assert fuzzy == {'startdate': date, 'enddate': date, 'nondate': ('State_Village_Tract_Boundaries ', ' lala'),
                         'date': ('10/12/01',)}
        with pytest.raises(ParserError):
            parse_date_range('lalala', '%d/%m/%Y')
        with pytest.raises(ParserError):
            parse_date_range('State_Village_Tract_Boundaries', fuzzy=dict())
        with pytest.raises(ParserError):
            fuzzy = dict()
            parse_date_range('Mon_State_Village_Tract_Boundaries', fuzzy=fuzzy)
        with pytest.raises(ParserError):
            parse_date_range('20/02')
        with pytest.raises(ParserError):
            parse_date_range('02/20')
        with pytest.raises(ParserError):
            parse_date_range('20/02', '%d/%m')

    def test_parse_date(self):
        assert parse_date('20/02/2013') == datetime(2013, 2, 20, 0, 0)
        assert parse_date('20/02/2013', '%d/%m/%Y') == datetime(2013, 2, 20, 0, 0)
        with pytest.raises(ParserError):
            parse_date('02/2013')
        with pytest.raises(ParserError):
            parse_date('02/2013', '%m/%Y')



# -*- coding: UTF-8 -*-
"""Logging Tests"""
from logging.handlers import SMTPHandler
from os.path import join

import pytest
from logging_tree import tree

from hdx.utilities.easy_logging import setup_logging, LoggingError

try:
    FILENOTFOUND_EXCTYPE = FileNotFoundError
except:
    FILENOTFOUND_EXCTYPE = IOError

class TestLogging:
    @pytest.fixture(scope='class')
    def logging_config_yaml(self):
        return join('tests', 'fixtures', 'config', 'logging_config.yml')

    @pytest.fixture(scope='class')
    def logging_config_json(self):
        return join('tests', 'fixtures', 'config', 'logging_config.json')

    @pytest.fixture(scope='class')
    def smtp_config_yaml(self):
        return join('tests', 'fixtures', 'config', 'smtp_config.yml')

    @pytest.fixture(scope='class')
    def smtp_config_json(self):
        return join('tests', 'fixtures', 'config', 'smtp_config.json')

    @staticmethod
    def get_root_handlers_names():
        return [x.name for x in tree()[1].handlers]

    @staticmethod
    def get_root_mail_handler_email():
        for handler in tree()[1].handlers:
            if isinstance(handler, SMTPHandler):
                return handler.toaddrs
        return None

    def test_setup_logging(self, logging_config_json, logging_config_yaml, smtp_config_json, smtp_config_yaml):
        with pytest.raises(FILENOTFOUND_EXCTYPE):
            setup_logging(smtp_config_json='NOT_EXIST')

        with pytest.raises(FILENOTFOUND_EXCTYPE):
            setup_logging(logging_config_json='NOT_EXIST', smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(FILENOTFOUND_EXCTYPE):
            setup_logging(logging_config_yaml='NOT_EXIST', smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(LoggingError):
            setup_logging(logging_config_yaml=logging_config_yaml, smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(LoggingError):
            setup_logging(logging_config_dict={'la': 'la'}, logging_config_json=logging_config_json,
                          smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(LoggingError):
            setup_logging(logging_config_dict={'la': 'la'}, logging_config_yaml=logging_config_yaml,
                          smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(LoggingError):
            setup_logging(logging_config_json=logging_config_json, logging_config_yaml=logging_config_yaml,
                          smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(LoggingError):
            setup_logging(smtp_config_dict={'la': 'la'}, smtp_config_json=smtp_config_json)

        with pytest.raises(LoggingError):
            setup_logging(smtp_config_dict={'la': 'la'}, smtp_config_yaml=smtp_config_yaml)

        with pytest.raises(LoggingError):
            setup_logging(smtp_config_json=smtp_config_json, smtp_config_yaml=smtp_config_yaml)

    def test_setup_logging_dict(self, smtp_config_yaml):
        handlers = ['error_file_handler', 'error_mail_handler']
        logging_config_dict = {'version': 1,
                               'handlers': {
                                   'error_file_handler': {
                                       'class': 'logging.FileHandler',
                                       'level': 'ERROR',
                                       'filename': 'errors.log',
                                       'encoding': 'utf8',
                                       'mode': 'w'
                                   },
                                   'error_mail_handler': {
                                       'class': 'logging.handlers.SMTPHandler',
                                       'level': 'CRITICAL',
                                       'mailhost': 'localhost',
                                       'fromaddr': 'noreply@localhost',
                                       'toaddrs': 'abc@abc.com',
                                       'subject': 'SCRAPER FAILED'
                                   }
                               },
                               'root': {
                                   'level': 'INFO',
                                   'handlers': handlers
                               }}
        setup_logging(logging_config_dict=logging_config_dict)
        actual_handler_names = self.get_root_handlers_names()
        for handler in handlers:
            assert handler in actual_handler_names
        assert 'abc@abc.com' in self.get_root_mail_handler_email()

    def test_setup_logging_json(self, logging_config_json):
        setup_logging(logging_config_json=logging_config_json)
        actual_handler_names = self.get_root_handlers_names()
        for handler in ['console', 'error_file_handler']:
            assert handler in actual_handler_names

    def test_setup_logging_yaml(self, monkeypatch, logging_config_yaml):
        setup_logging(logging_config_yaml=logging_config_yaml)
        actual_handler_names = self.get_root_handlers_names()
        for handler in ['console', 'error_mail_handler']:
            assert handler in actual_handler_names
        assert 'abc@abc.com' in self.get_root_mail_handler_email()

    def test_setup_logging_env(self, monkeypatch, logging_config_yaml):
        monkeypatch.setenv('LOG_FILE_ONLY', 'true')
        setup_logging(logging_config_yaml=logging_config_yaml)
        actual_handler_names = self.get_root_handlers_names()
        assert 'console' not in actual_handler_names
        monkeypatch.setenv('LOG_FILE_ONLY', 'false')
        setup_logging(logging_config_yaml=logging_config_yaml)
        actual_handler_names = self.get_root_handlers_names()
        assert 'console' in actual_handler_names

    def test_setup_logging_smtp_dict(self):
        smtp_config_dict = {'handlers': {'error_mail_handler': {'toaddrs': 'lalala@la.com', 'subject': 'lala'}}}
        setup_logging(smtp_config_dict=smtp_config_dict)
        actual_handler_names = self.get_root_handlers_names()
        for handler in ['console', 'error_file_handler', 'error_mail_handler']:
            assert handler in actual_handler_names
        assert 'lalala@la.com' in self.get_root_mail_handler_email()

    def test_setup_logging_smtp_json(self, smtp_config_json):
        setup_logging(smtp_config_json=smtp_config_json)
        actual_handler_names = self.get_root_handlers_names()
        for handler in ['console', 'error_file_handler', 'error_mail_handler']:
            assert handler in actual_handler_names
        assert '123@123.com' in self.get_root_mail_handler_email()

    def test_setup_logging_smtp_yaml(self, smtp_config_yaml):
        setup_logging(smtp_config_yaml=smtp_config_yaml)
        actual_handler_names = self.get_root_handlers_names()
        for handler in ['console', 'error_file_handler', 'error_mail_handler']:
            assert handler in actual_handler_names
        assert 'abc@abc.com' in self.get_root_mail_handler_email()



# -*- coding: UTF-8 -*-
"""Configuration of logging"""
import logging.config
import os
from typing import Any

from hdx.utilities.dictandlist import merge_dictionaries
from hdx.utilities.loader import load_yaml, load_json
from hdx.utilities.path import script_dir_plus_file


class LoggingError(Exception):
    pass


def setup_logging(**kwargs):
    # type: (Any) -> None
    """Setup logging configuration

    Args:
        **kwargs: See below
        logging_config_dict (dict): Logging configuration dictionary OR
        logging_config_json (str): Path to JSON Logging configuration OR
        logging_config_yaml (str): Path to YAML Logging configuration. Defaults to internal logging_configuration.yml.
        smtp_config_dict (dict): Email Logging configuration dictionary if using default logging configuration OR
        smtp_config_json (str): Path to JSON Email Logging configuration if using default logging configuration OR
        smtp_config_yaml (str): Path to YAML Email Logging configuration if using default logging configuration

    Returns:
        None
    """
    smtp_config_found = False
    smtp_config_dict = kwargs.get('smtp_config_dict', None)
    if smtp_config_dict:
        smtp_config_found = True
        print('Loading smtp configuration customisations from dictionary')

    smtp_config_json = kwargs.get('smtp_config_json', '')
    if smtp_config_json:
        if smtp_config_found:
            raise LoggingError('More than one smtp configuration file given!')
        smtp_config_found = True
        print('Loading smtp configuration customisations from: %s' % smtp_config_json)
        smtp_config_dict = load_json(smtp_config_json)

    smtp_config_yaml = kwargs.get('smtp_config_yaml', '')
    if smtp_config_yaml:
        if smtp_config_found:
            raise LoggingError('More than one smtp configuration file given!')
        smtp_config_found = True
        print('Loading smtp configuration customisations from: %s' % smtp_config_yaml)
        smtp_config_dict = load_yaml(smtp_config_yaml)

    logging_smtp_config_dict = None
    logging_config_found = False
    logging_config_dict = kwargs.get('logging_config_dict', None)
    if logging_config_dict:
        logging_config_found = True
        print('Loading logging configuration from dictionary')

    logging_config_json = kwargs.get('logging_config_json', '')
    if logging_config_json:
        if logging_config_found:
            raise LoggingError('More than one logging configuration file given!')
        logging_config_found = True
        print('Loading logging configuration from: %s' % logging_config_json)
        logging_config_dict = load_json(logging_config_json)

    logging_config_yaml = kwargs.get('logging_config_yaml', '')
    if logging_config_found:
        if logging_config_yaml:
            raise LoggingError('More than one logging configuration file given!')
    else:
        if not logging_config_yaml:
            print('No logging configuration parameter. Using default.')
            logging_config_yaml = script_dir_plus_file('logging_configuration.yml', setup_logging)
            if smtp_config_found:
                logging_smtp_config_yaml = script_dir_plus_file('logging_smtp_configuration.yml', setup_logging)
                print('Loading base SMTP logging configuration from: %s' % logging_smtp_config_yaml)
                logging_smtp_config_dict = load_yaml(logging_smtp_config_yaml)
        print('Loading logging configuration from: %s' % logging_config_yaml)
        logging_config_dict = load_yaml(logging_config_yaml)

    if smtp_config_found:
        if logging_smtp_config_dict:
            logging_config_dict = merge_dictionaries([logging_config_dict, logging_smtp_config_dict, smtp_config_dict])
        else:
            raise LoggingError('SMTP logging configuration file given but not using default logging configuration!')
    file_only = os.getenv('LOG_FILE_ONLY')
    if file_only is not None and file_only.lower() not in ['false', 'f', 'n', 'no', '0']:
        root = logging_config_dict.get('root')
        if root is not None:
            handlers = root.get('handlers', list())
            for i, handler in enumerate(handlers):
                if handler.lower() == 'console':
                    del handlers[i]
                    break

    logging.config.dictConfig(logging_config_dict)



# -*- coding: UTF-8 -*-
"""Saver Tests"""
from collections import OrderedDict
from os.path import join

import pytest

from hdx.utilities.compare import assert_files_same
from hdx.utilities.saver import save_yaml, save_json


class TestLoader:
    yaml_to_write = OrderedDict([('hdx_prod_site', OrderedDict([('url', 'https://data.humdata.org/'),
                                                                ('username', None), ('password', None)])),
                                 ('hdx_test_site', OrderedDict([('url', 'https://test-data.humdata.org/'),
                                                                ('username', 'lala'), ('password', 'lalala')])),
                                 ('dataset', OrderedDict([('required_fields', ['name', 'title', 'dataset_date'])])),
                                 (
                                     'resource',
                                     OrderedDict([('required_fields', ['package_id', 'name', 'description'])])),
                                 ('showcase', OrderedDict([('required_fields', ['name', 'title'])])),
                                 ('param_1', 'ABC')])

    json_to_write = OrderedDict([('hdx_prod_site', OrderedDict([('url', 'https://data.humdata.org/'),
                                                                ('username', None), ('password', None)])),
                                 ('hdx_test_site', OrderedDict([('url', 'https://test-data.humdata.org/'),
                                                                ('username', 'tumteetum'),
                                                                ('password', 'tumteetumteetum')])),
                                 ('dataset', OrderedDict([('required_fields', ['name', 'dataset_date'])])),
                                 ('resource', OrderedDict([('required_fields', ['name', 'description'])])),
                                 ('showcase', OrderedDict([('required_fields', ['name'])])), ('my_param', 'abc')])

    @pytest.fixture(scope='session')
    def saverfolder(self, fixturesfolder):
        return join(fixturesfolder, 'saver')

    @pytest.mark.parametrize('filename,pretty,sortkeys', [
        ('pretty-false_sortkeys-false.yaml', False, False),
        ('pretty-false_sortkeys-true.yaml', False, True),
        ('pretty-true_sortkeys-false.yaml', True, False),
        ('pretty-true_sortkeys-true.yaml', True, True),
    ])
    def test_save_yaml(self, tmpdir, saverfolder, filename, pretty, sortkeys):
        test_path = join(str(tmpdir), filename)
        ref_path = join(saverfolder, filename)
        save_yaml(TestLoader.yaml_to_write, test_path, pretty=pretty, sortkeys=sortkeys)
        assert_files_same(ref_path, test_path)

    @pytest.mark.parametrize('filename,pretty,sortkeys', [
        ('pretty-false_sortkeys-false.json', False, False),
        ('pretty-false_sortkeys-true.json', False, True),
        ('pretty-true_sortkeys-false.json', True, False),
        ('pretty-true_sortkeys-true.json', True, True),
    ])
    def test_save_json(self, tmpdir, saverfolder, filename, pretty, sortkeys):
        test_path = join(str(tmpdir), filename)
        ref_path = join(saverfolder, filename)
        save_json(TestLoader.json_to_write, test_path, pretty=pretty, sortkeys=sortkeys)
        assert_files_same(ref_path, test_path)



# -*- coding: UTF-8 -*-
"""Path Utility Tests"""
from os.path import join, exists
from shutil import rmtree
from tempfile import gettempdir

import pytest

from hdx.utilities.path import get_temp_dir, temp_dir, progress_storing_tempdir


class TestPath:
    @pytest.fixture(scope='class')
    def mytestdir(self):
        return join('haha', 'lala')

    def test_get_temp_dir(self, monkeypatch, mytestdir):
        assert get_temp_dir() == gettempdir()
        assert get_temp_dir('TEST') == join(gettempdir(), 'TEST')
        monkeypatch.setenv('TEMP_DIR', mytestdir)
        assert get_temp_dir() == mytestdir
        monkeypatch.delenv('TEMP_DIR')

    def test_temp_dir(self, monkeypatch, mytestdir):
        monkeypatch.setenv('TEMP_DIR', mytestdir)
        with temp_dir() as tempdir:
            assert tempdir == mytestdir
        monkeypatch.delenv('TEMP_DIR')

        tempfolder = 'papa'
        expected_dir = join(gettempdir(), tempfolder)

        with temp_dir(tempfolder) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is False
        try:
            with temp_dir(tempfolder) as tempdir:
                assert tempdir == expected_dir
                raise ValueError('Fail!')
        except:
            pass
        assert exists(tempdir) is False

        with temp_dir(tempfolder, delete_on_success=True, delete_on_failure=True) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is False
        try:
            with temp_dir(tempfolder, delete_on_success=True, delete_on_failure=True) as tempdir:
                assert tempdir == expected_dir
                raise ValueError('Fail!')
        except:
            pass
        assert exists(tempdir) is False

        with temp_dir(tempfolder, delete_on_success=False, delete_on_failure=False) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is True
        rmtree(tempdir)
        try:
            with temp_dir(tempfolder, delete_on_success=False, delete_on_failure=False) as tempdir:
                assert tempdir == expected_dir
                raise ValueError('Fail!')
        except:
            pass
        assert exists(tempdir) is True

        with temp_dir(tempfolder, delete_on_success=True, delete_on_failure=False) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is False
        try:
            with temp_dir(tempfolder, delete_on_success=True, delete_on_failure=False) as tempdir:
                assert tempdir == expected_dir
                raise ValueError('Fail!')
        except:
            pass
        assert exists(tempdir) is True
        rmtree(tempdir)

        with temp_dir(tempfolder, delete_on_success=False, delete_on_failure=True) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is True
        rmtree(tempdir)
        try:
            with temp_dir(tempfolder, delete_on_success=False, delete_on_failure=True) as tempdir:
                assert tempdir == expected_dir
                raise ValueError('Fail!')
        except:
            pass
        assert exists(tempdir) is False

    def test_progress_storing_tempdir(self, monkeypatch):
        tempfolder = 'papa'
        expected_dir = join(gettempdir(), tempfolder)
        iterator = [{'iso3': 'AFG', 'name': 'Afghanistan'}, {'iso3': 'SDN', 'name': 'Sudan'},
                    {'iso3': 'YEM', 'name': 'Yemen'}, {'iso3': 'ZAM', 'name': 'Zambia'}]
        result = list()
        for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
            assert exists(tempdir) is True
            assert tempdir == expected_dir
            result.append(nextdict)
        assert result == iterator
        assert exists(expected_dir) is False

        monkeypatch.setenv('WHERETOSTART', 'SDN')
        result = list()
        for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
            assert exists(tempdir) is True
            assert tempdir == expected_dir
            result.append(nextdict)
        assert result == iterator[1:]
        assert exists(expected_dir) is False
        monkeypatch.delenv('WHERETOSTART')

        try:
            for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
                if nextdict['iso3'] == 'YEM':
                    raise ValueError('Problem!')
        except:
            pass
        assert exists(expected_dir) is True
        result = list()
        for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
            assert exists(tempdir) is True
            assert tempdir == expected_dir
            result.append(nextdict)
        assert result == iterator[2:]
        assert exists(expected_dir) is False

        try:
            for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
                if nextdict['iso3'] == 'YEM':
                    raise ValueError('Problem!')
        except:
            pass
        assert exists(expected_dir) is True
        monkeypatch.setenv('WHERETOSTART', 'RESET')
        result = list()
        for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
            assert exists(tempdir) is True
            assert tempdir == expected_dir
            result.append(nextdict)
        assert result == iterator
        assert exists(expected_dir) is False
        monkeypatch.delenv('WHERETOSTART')

        try:
            for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
                if nextdict['iso3'] == 'YEM':
                    raise ValueError('Problem!')
        except:
            pass
        assert exists(expected_dir) is True
        monkeypatch.setenv('WHERETOSTART', 'SDN')
        result = list()
        for tempdir, nextdict in progress_storing_tempdir(tempfolder, iterator, 'iso3'):
            assert exists(tempdir) is True
            assert tempdir == expected_dir
            result.append(nextdict)
        assert result == iterator[1:]
        assert exists(expected_dir) is False
        monkeypatch.delenv('WHERETOSTART')



# -*- coding: UTF-8 -*-
"""Encoding Utility Tests"""
from hdx.utilities.encoding import str_to_base64, base64_to_str


class TestEncoding:
    def test_base64(self):
        string = 'hello'
        result = str_to_base64(string)
        assert result == 'aGVsbG8='
        assert base64_to_str(result) == string



# -*- coding: utf-8 -*-
"""Loading utilities for YAML, JSON etc."""

import json
from collections import OrderedDict
from io import open
from os import linesep
from typing import List, Dict, Optional

import yaml
import yamlloader

from hdx.utilities.dictandlist import merge_two_dictionaries, merge_dictionaries


class LoadError(Exception):
    pass


def load_and_merge_yaml(paths, encoding='utf-8'):
    # type: (List[str], str) -> Dict
    """Load multiple YAML files and merge into one dictionary

    Args:
        paths (List[str]): Paths to YAML files
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Dict: Dictionary of merged YAML files

    """
    configs = [load_yaml(path, encoding=encoding) for path in paths]
    return merge_dictionaries(configs)


def load_and_merge_json(paths, encoding='utf-8'):
    # type: (List[str], str) -> Dict
    """Load multiple JSON files and merge into one dictionary

    Args:
        paths (List[str]): Paths to JSON files
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Dict: Dictionary of merged JSON files

    """
    configs = [load_json(path, encoding=encoding) for path in paths]
    return merge_dictionaries(configs)


def load_yaml_into_existing_dict(data, path, encoding='utf-8'):
    # type: (dict, str, str) -> Dict
    """Merge YAML file into existing dictionary

    Args:
        data (dict): Dictionary to merge into
        path (str): YAML file to load and merge
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Dict: YAML file merged into dictionary
    """
    yamldict = load_yaml(path, encoding=encoding)
    return merge_two_dictionaries(data, yamldict)


def load_json_into_existing_dict(data, path, encoding='utf-8'):
    # type: (dict, str, str) -> Dict
    """Merge JSON file into existing dictionary

    Args:
        data (dict): Dictionary to merge into
        path (str): JSON file to load and merge
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        dict: JSON file merged into dictionary
    """
    jsondict = load_json(path, encoding=encoding)
    return merge_two_dictionaries(data, jsondict)


def load_yaml(path, encoding='utf-8'):
    # type: (str, str) -> OrderedDict
    """Load YAML file into an ordered dictionary

    Args:
        path (str): Path to YAML file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        OrderedDict: Ordered dictionary containing loaded YAML file
    """
    with open(path, 'r', encoding=encoding) as f:
        yamldict = yaml.load(f.read(), Loader=yamlloader.ordereddict.CSafeLoader)
    if not yamldict:
        raise (LoadError('YAML file: %s is empty!' % path))
    return yamldict


def load_json(path, encoding='utf-8'):
    # type: (str, str) -> OrderedDict
    """Load JSON file into an ordered dictionary

    Args:
        path (str): Path to JSON file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        OrderedDict: Ordered dictionary containing loaded JSON file
    """
    with open(path, 'r', encoding=encoding) as f:
        jsondict = json.loads(f.read(), object_pairs_hook=OrderedDict)
    if not jsondict:
        raise (LoadError('JSON file: %s is empty!' % path))
    return jsondict


def load_file_to_str(path, encoding='utf-8', strip=False, replace_newlines=None):
    # type: (str, str, bool, Optional[str]) -> str
    """
    Load file into a string removing newlines

    Args:
        path (str): Path to file
        encoding (str): Encoding of file. Defaults to utf-8.
        strip (bool): Whether to strip whitespace from start and end. Defaults to False.
        replace_newlines (Optional[str]): String with which tp replace newlines. Defaults to None (don't replace).

    Returns:
        str: String contents of file

    """
    with open(path, 'r', encoding=encoding) as f:
        string = f.read()
        if replace_newlines is not None:
            string = string.replace(linesep, replace_newlines)
        if strip:
            string = string.strip()
    if not string:
        raise LoadError('%s file is empty!' % path)
    return string



# -*- coding: UTF-8 -*-
"""User Agent Tests"""
from os.path import join

import pytest

from hdx.utilities.loader import LoadError
from hdx.utilities.useragent import UserAgent, UserAgentError
from hdx.utilities.version import get_utils_version


class TestUserAgent:
    @pytest.fixture(scope='class')
    def user_agent_config_yaml(self, configfolder):
        return join(configfolder, 'user_agent_config.yml')

    @pytest.fixture(scope='class')
    def user_agent_config2_yaml(self, configfolder):
        return join(configfolder, 'user_agent_config2.yml')

    @pytest.fixture(scope='class')
    def user_agent_config3_yaml(self, configfolder):
        return join(configfolder, 'user_agent_config3.yml')

    @pytest.fixture(scope='class')
    def empty_yaml(self, configfolder):
        return join(configfolder, 'empty.yml')

    @pytest.fixture(scope='class')
    def user_agent_config_wrong_yaml(self, configfolder):
        return join(configfolder, 'user_agent_config_wrong.yml')

    def test_user_agent(self, monkeypatch, user_agent_config_yaml, user_agent_config2_yaml, user_agent_config3_yaml,
                        empty_yaml, user_agent_config_wrong_yaml):
        version = get_utils_version()
        assert UserAgent.get(
            user_agent_config_yaml=user_agent_config_yaml) == 'lala:HDXPythonUtilities/%s-myua' % version
        assert UserAgent.get(
            user_agent_config_yaml=user_agent_config2_yaml) == 'HDXPythonUtilities/%s-myuseragent' % version
        assert UserAgent.get(user_agent_config_yaml=user_agent_config3_yaml,
                             user_agent_lookup='lookup') == 'HDXPythonUtilities/%s-mylookupagent' % version
        assert UserAgent.get(user_agent='my_ua', preprefix='papa') == 'papa:HDXPythonUtilities/%s-my_ua' % version
        UserAgent.set_global(user_agent_config_yaml=user_agent_config3_yaml, user_agent_lookup='lookup2')
        assert UserAgent.get() == 'HDXPythonUtilities/%s-mylookupagent2' % version
        UserAgent.clear_global()
        with pytest.raises(UserAgentError):
            UserAgent.get(user_agent_config_yaml=user_agent_config3_yaml, user_agent_lookup='fail')
        with pytest.raises(LoadError):
            UserAgent.get(user_agent_config_yaml=empty_yaml)
        with pytest.raises(UserAgentError):
            UserAgent.get(user_agent_config_yaml=user_agent_config_wrong_yaml)
        with pytest.raises(UserAgentError):
            UserAgent.get(user_agent_config_yaml='')
        with pytest.raises(UserAgentError):
            UserAgent.get()
        with pytest.raises(UserAgentError):
            UserAgent._load(prefix='', user_agent_config_yaml='')
        my_user_agent = 'lala'
        monkeypatch.setenv('USER_AGENT', my_user_agent)
        assert UserAgent.get() == 'HDXPythonUtilities/%s-%s' % (version, my_user_agent)
        my_preprefix = 'haha'
        monkeypatch.setenv('PREPREFIX', my_preprefix)
        assert UserAgent.get() == '%s:HDXPythonUtilities/%s-%s' % (my_preprefix, version, my_user_agent)
        my_prefix = 'HDXPythonLibrary/%s' % version
        assert UserAgent.get(prefix=my_prefix) == '%s:%s-%s' % (my_preprefix, my_prefix, my_user_agent)
        UserAgent.clear_global()



# -*- coding: utf-8 -*-
"""File compare utilities"""
import difflib
from os import linesep
from typing import List


def compare_files(path1, path2):
    # type: (str, str) -> List[str]
    """Returns the delta between two files using -, ?, + format excluding
    lines that are the same

    Args:
        path1 (str): Path to first file
        path2 (str): Path to second file

    Returns:
        List[str]: Delta between the two files

    """
    diff = difflib.ndiff(open(path1).read().splitlines(), open(path2).read().splitlines())
    return [x for x in diff if x[0] in ['-', '+', '?']]


def assert_files_same(path1, path2):
    # type: (str, str) -> None
    """Asserts that two files are the same and returns delta using
    -, ?, + format if not

    Args:
        path1 (str): Path to first file
        path2 (str): Path to second file

    Returns:
        None

    """
    difflines = compare_files(path1, path2)
    assert len(difflines) == 0, linesep.join([linesep] + difflines)



# -*- coding: utf-8 -*-
"""Session utilities for urls"""
import logging
import os
from typing import Any, Optional

import requests
from basicauth import decode
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from hdx.utilities.loader import load_file_to_str, load_json, load_yaml
from hdx.utilities.useragent import UserAgent

logger = logging.getLogger(__name__)


class SessionError(Exception):
    pass


def get_session(user_agent=None, user_agent_config_yaml=None, user_agent_lookup=None, use_env=True, **kwargs):
    # type: (Optional[str], Optional[str], Optional[str], bool, Any) -> requests.Session
    """Set up and return Session object that is set up with retrying. Requires either global user agent to be set or
    appropriate user agent parameter(s) to be completed. If the EXTRA_PARAMS or BASIC_AUTH environment variable is
    supplied, the extra_params* parameters will be ignored.

    Args:
        user_agent (Optional[str]): User agent string. HDXPythonUtilities/X.X.X- is prefixed.
        user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
        user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.
        use_env (bool): Whether to read environment variables. Defaults to True.
        **kwargs: See below
        auth (Tuple[str, str]): Authorisation information in tuple form (user, pass) OR
        basic_auth (str): Authorisation information in basic auth string form (Basic xxxxxxxxxxxxxxxx) OR
        basic_auth_file (str): Path to file containing authorisation information in basic auth string form (Basic xxxxxxxxxxxxxxxx)
        extra_params_dict (Dict): Extra parameters to put on end of url as a dictionary OR
        extra_params_json (str): Path to JSON file containing extra parameters to put on end of url OR
        extra_params_yaml (str): Path to YAML file containing extra parameters to put on end of url
        extra_params_lookup (str): Lookup key for parameters. If not given assumes parameters are at root of the dict.
        status_forcelist (iterable): HTTP statuses for which to force retry. Defaults to [429, 500, 502, 503, 504].
        method_whitelist (iterable): HTTP methods for which to force retry. Defaults t0 frozenset(['GET']).
    """
    s = requests.Session()

    ua = kwargs.get('full_agent')
    if not ua:
        ua = UserAgent.get(user_agent, user_agent_config_yaml, user_agent_lookup, **kwargs)
    s.headers['User-Agent'] = ua

    extra_params_found = False
    extra_params_dict = None
    auth_found = False
    basic_auth = None
    if use_env:
        basic_auth = os.getenv('BASIC_AUTH')
        if basic_auth:
            logger.info('Loading authorisation from basic_auth environment variable')
            auth_found = True
        extra_params = os.getenv('EXTRA_PARAMS')
        if extra_params:
            if '=' in extra_params:
                extra_params_dict = dict()
                logger.info('Loading extra parameters from environment variable')
                for extra_param in extra_params.split(','):
                    key, value = extra_param.split('=')
                    extra_params_dict[key] = value
            extra_params_found = True
    if not auth_found and not extra_params_found:
        # only do this if basic authorisation and extra params env vars not supplied
        extra_params_dict = kwargs.get('extra_params_dict')
        if extra_params_dict:
            extra_params_found = True
            logger.info('Loading extra parameters from dictionary')

        extra_params_json = kwargs.get('extra_params_json', '')
        if extra_params_json:
            if extra_params_found:
                raise SessionError('More than one set of extra parameters given!')
            extra_params_found = True
            logger.info('Loading extra parameters from: %s' % extra_params_json)
            extra_params_dict = load_json(extra_params_json)

        extra_params_yaml = kwargs.get('extra_params_yaml', '')
        if extra_params_yaml:
            if extra_params_found:
                raise SessionError('More than one set of extra parameters given!')
            logger.info('Loading extra parameters from: %s' % extra_params_yaml)
            extra_params_dict = load_yaml(extra_params_yaml)
        extra_params_lookup = kwargs.get('extra_params_lookup')
        if extra_params_lookup:
            extra_params_dict = extra_params_dict.get(extra_params_lookup)
            if extra_params_dict is None:
                raise SessionError('%s does not exist in extra_params!' % extra_params_lookup)

    if not auth_found:
        basic_auth = kwargs.get('basic_auth')
        if basic_auth:
            logger.info('Loading authorisation from basic_auth argument')
            auth_found = True
        elif extra_params_dict:
            bauth = extra_params_dict.get('basic_auth')
            if bauth:
                basic_auth = bauth
                logger.info('Loading authorisation from basic_auth parameter')
                auth_found = True
                del extra_params_dict['basic_auth']

    s.params = extra_params_dict
    auth = kwargs.get('auth')
    if auth:
        if auth_found:
            raise SessionError('More than one authorisation given!')
        logger.info('Loading authorisation from auth argument')
        auth_found = True
    basic_auth_file = kwargs.get('basic_auth_file')
    if basic_auth_file:
        if auth_found:
            raise SessionError('More than one authorisation given!')
        logger.info('Loading authorisation from: %s' % basic_auth_file)
        basic_auth = load_file_to_str(basic_auth_file, strip=True)
    if basic_auth:
        auth = decode(basic_auth)
    s.auth = auth

    status_forcelist = kwargs.get('status_forcelist', [429, 500, 502, 503, 504])
    method_whitelist = kwargs.get('method_whitelist', frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']))

    retries = Retry(total=5, backoff_factor=0.4, status_forcelist=status_forcelist, method_whitelist=method_whitelist,
                    raise_on_redirect=True,
                    raise_on_status=True)
    s.mount('http://', HTTPAdapter(max_retries=retries, pool_connections=100, pool_maxsize=100))
    s.mount('https://', HTTPAdapter(max_retries=retries, pool_connections=100, pool_maxsize=100))
    return s



# -*- coding: utf-8 -*-
"""Utility class to simplify sending emails"""

import logging
import smtplib
from os.path import join, expanduser
from typing import Optional, Any, List

from email_validator import validate_email
from six.moves.email_mime_multipart import MIMEMultipart
from six.moves.email_mime_text import MIMEText

from hdx.utilities.loader import load_yaml, load_json

logger = logging.getLogger(__name__)


class EmailConfigurationError(Exception):
    pass


class Email:
    """
    Emailer utility. Parameters in dictionary or file (eg. yaml below):
     | connection_type: "ssl"   ("ssl" for smtp ssl or "lmtp", otherwise basic smtp is assumed)
     | host: "localhost"
     | port: 123
     | local_hostname: "mycomputer.fqdn.com"
     | timeout: 3
     | username: "user"
     | password: "pass"
     | sender: "a@b.com"    (if not supplied username is used as sender)

    Args:
        **kwargs: See below
        email_config_dict (dict): HDX configuration dictionary OR
        email_config_json (str): Path to JSON HDX configuration OR
        email_config_yaml (str): Path to YAML HDX configuration. Defaults to ~/hdx_email_configuration.yml.
    """

    default_email_config_yaml = join(expanduser('~'), 'hdx_email_configuration.yml')

    def __init__(self, **kwargs):
        # type: (Any) -> None
        email_config_found = False
        email_config_dict = kwargs.get('email_config_dict', None)
        if email_config_dict:
            email_config_found = True
            logger.info('Loading email configuration from dictionary')

        email_config_json = kwargs.get('email_config_json', '')
        if email_config_json:
            if email_config_found:
                raise EmailConfigurationError('More than one email configuration file given!')
            email_config_found = True
            logger.info('Loading email configuration from: %s' % email_config_json)
            email_config_dict = load_json(email_config_json)

        email_config_yaml = kwargs.get('email_config_yaml', None)
        if email_config_found:
            if email_config_yaml:
                raise EmailConfigurationError('More than one email configuration file given!')
        else:
            if not email_config_yaml:
                logger.info('No email configuration parameter. Using default email configuration path.')
                email_config_yaml = Email.default_email_config_yaml
            logger.info('Loading email configuration from: %s' % email_config_yaml)
            email_config_dict = load_yaml(email_config_yaml)

        self.connection_type = email_config_dict.get('connection_type', 'smtp')
        self.host = email_config_dict.get('host', '')
        self.port = email_config_dict.get('port', 0)
        self.local_hostname = email_config_dict.get('local_hostname')
        self.timeout = email_config_dict.get('timeout')
        self.source_address = email_config_dict.get('source_address')
        self.username = email_config_dict.get('username')
        self.password = email_config_dict.get('password')
        self.sender = email_config_dict.get('sender', self.username)
        self.server = None

    def __enter__(self):
        # type: () -> 'Email'
        """
        Return Email object for with statement

        Returns:
            None

        """
        return self

    def __exit__(self, *args):
        # type: (Any) -> None
        """
        Close Email object for end of with statement

        Args:
            *args: Not used

        Returns:
            None

        """
        pass

    def connect(self):
        # type: () -> None
        """
        Connect to server

        Returns:
            None

        """
        if self.connection_type.lower() == 'ssl':
            self.server = smtplib.SMTP_SSL(host=self.host, port=self.port, local_hostname=self.local_hostname,
                                           timeout=self.timeout, source_address=self.source_address)
        elif self.connection_type.lower() == 'lmtp':
            self.server = smtplib.LMTP(host=self.host, port=self.port, local_hostname=self.local_hostname,
                                       source_address=self.source_address)
        else:
            self.server = smtplib.SMTP(host=self.host, port=self.port, local_hostname=self.local_hostname,
                                       timeout=self.timeout, source_address=self.source_address)
        self.server.login(self.username, self.password)

    def close(self):
        # type: () -> None
        """
        Close connection to email server

        Returns:
            None

        """
        self.server.quit()

    @staticmethod
    def get_normalised_emails(recipients):
        # type: (List[str]) -> List[str]
        """
        Get list of normalised emails

        Args:
            recipients (List[str]): Email recipient

        Returns:
            List[str]: Normalised emails

        """
        normalised_recipients = list()
        for recipient in recipients:
            v = validate_email(recipient, check_deliverability=True)  # validate and get info
            normalised_recipients.append(v['email'])  # replace with normalized form
        return normalised_recipients

    def send(self, recipients, subject, text_body, html_body=None, sender=None, cc=None, bcc=None, **kwargs):
        # type: (List[str], str, str, Optional[str], Optional[str], Optional[List[str]], Optional[List[str]], Any) -> None
        """
        Send email

        Args:
            recipients (List[str]): Email recipient
            subject (str): Email subject
            text_body (str): Plain text email body
            html_body (Optional[str]): HTML email body
            sender (Optional[str]): Email sender. Defaults to global sender.
            cc (Optional[List[str]]): Email cc. Defaults to None.
            bcc (Optional[List[str]]): Email cc. Defaults to None.
            **kwargs: See below
            mail_options (list): Mail options (see smtplib documentation)
            rcpt_options (list): Recipient options (see smtplib documentation)

        Returns:
            None

        """
        if sender is None:
            sender = self.sender
        v = validate_email(sender, check_deliverability=False)  # validate and get info
        sender = v['email']  # replace with normalized form

        if html_body is not None:
            msg = MIMEMultipart('alternative')
            part1 = MIMEText(text_body, 'plain')
            part2 = MIMEText(html_body, 'html')
            msg.attach(part1)
            msg.attach(part2)
        else:
            msg = MIMEText(text_body)
        msg['Subject'] = subject
        msg['From'] = sender

        normalised_recipients = self.get_normalised_emails(recipients)
        msg['To'] = ', '.join(normalised_recipients)

        if cc is not None:
            normalised_cc = self.get_normalised_emails(cc)
            msg['Cc'] = ', '.join(normalised_cc)
            normalised_recipients.extend(normalised_cc)

        if bcc is not None:
            normalised_bcc = self.get_normalised_emails(bcc)
            normalised_recipients.extend(normalised_bcc)

        # Perform operations via server
        self.connect()
        self.server.sendmail(sender, normalised_recipients, msg.as_string(), **kwargs)
        self.close()



# -*- coding: UTF-8 -*-
"""UUID Tests"""

from hdx.utilities import is_valid_uuid


class TestUUID:
    def test_is_valid_uuid(self):
        assert is_valid_uuid('a232') is False
        assert is_valid_uuid(True) is False
        assert is_valid_uuid(123) is False
        assert is_valid_uuid('') is False
        assert is_valid_uuid('jpsmith') is False
        assert is_valid_uuid('c9bf9e57-1685-4c89-bafb-ff5af830be8a') is True



# -*- coding: utf-8 -*-
"""Database utilities"""
import logging
import time
from typing import Dict, Any, Optional, Union

import psycopg2
from six.moves.urllib.parse import urlsplit
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declared_attr, declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
from sqlalchemy.pool import NullPool
from sshtunnel import SSHTunnelForwarder

logger = logging.getLogger(__name__)


class HDXBase(object):
    @declared_attr
    def __tablename__(cls):
        return '%ss' % cls.__name__.lower()


Base = declarative_base(cls=HDXBase)


class Database(object):
    """Database helper class to handle ssh tunnels, waiting for PostgreSQL to be up etc.

    Args:
        database (Optional[str]): Database name
        host (Optional[str]): Host where database is located
        port (Union[int, str, None]): Database port
        username (Optional[str]): Username to log into database
        password (Optional[str]): Password to log into database
        driver (str): Database driver. Defaults to 'postgres'.
        **kwargs: See below
        ssh_host (str): SSH host (the server to connect to)
        ssh_port (int): SSH port. Defaults to 22.
        ssh_username (str): SSH username
        ssh_password (str): SSH password
        ssh_private_key: Location of SSH private key (instead of ssh_password)
        For more advanced usage, see SSHTunnelForwarder documentation.

    """

    def __init__(self, database=None, host=None, port=None, username=None, password=None, driver='postgres', **kwargs):
        # type: (Optional[str], Optional[str], Union[int, str, None], Optional[str], Optional[str], str, Any) -> None
        if port is not None:
            port = int(port)
        if len(kwargs) != 0:
            ssh_host = kwargs['ssh_host']
            del kwargs['ssh_host']
            ssh_port = kwargs.get('ssh_port')
            if ssh_port is not None:
                ssh_port = int(ssh_port)
                del kwargs['ssh_port']
            else:
                ssh_port = 22
            self.server = SSHTunnelForwarder((ssh_host, ssh_port), remote_bind_address=(host, port), **kwargs)
            self.server.start()
            host = self.server.local_bind_host
            port = self.server.local_bind_port
        else:
            self.server = None
        if driver == 'postgres':
            Database.wait_for_postgres(database, host, port, username, password)
        db_url = self.get_sqlalchemy_url(database, host, port, username, password, driver=driver)
        self.session = self.get_session(db_url)

    def __enter__(self):
        # type: () -> Session
        return self.session

    def __exit__(self, exc_type, exc_value, traceback):
        # type: (Any, Any, Any) -> None
        self.session.close()
        if self.server is not None:
            self.server.stop()

    @staticmethod
    def get_session(db_url):
        # type: (str) -> Session
        """Gets SQLAlchemy session given url. Your tables must inherit
        from Base in hdx.utilities.database.

        Args:
            db_url (str): SQLAlchemy url

        Returns:
            sqlalchemy.orm.session.Session: SQLAlchemy session
        """
        engine = create_engine(db_url, poolclass=NullPool, echo=False)
        Session = sessionmaker(bind=engine)
        Base.metadata.create_all(engine)
        return Session()

    @staticmethod
    def get_params_from_sqlalchemy_url(db_url):
        # type: (str) -> Dict[str,Any]
        """Gets PostgreSQL database connection parameters from SQLAlchemy url

        Args:
            db_url (str): SQLAlchemy url

        Returns:
            Dict[str,Any]: Dictionary of database connection parameters
        """
        result = urlsplit(db_url)
        return {'database': result.path[1:], 'host': result.hostname, 'port': result.port,
                'username': result.username, 'password': result.password, 'driver': result.scheme}

    @staticmethod
    def get_sqlalchemy_url(database=None, host=None, port=None, username=None, password=None, driver='postgres'):
        # type: (Optional[str], Optional[str], Union[int, str, None], Optional[str], Optional[str], str) -> str
        """Gets SQLAlchemy url from database connection parameters

        Args:
            database (Optional[str]): Database name
            host (Optional[str]): Host where database is located
            port (Union[int, str, None]): Database port
            username (Optional[str]): Username to log into database
            password (Optional[str]): Password to log into database
            driver (str): Database driver. Defaults to 'postgres'.

        Returns:
            db_url (str): SQLAlchemy url
        """
        strings = ['%s://' % driver]
        if username:
            strings.append(username)
            if password:
                strings.append(':%s@' % password)
            else:
                strings.append('@')
        if host:
            strings.append(host)
        if port is not None:
            strings.append(':%d' % int(port))
        if database:
            strings.append('/%s' % database)
        return ''.join(strings)

    @staticmethod
    def wait_for_postgres(database, host, port, username, password):
        # type: (Optional[str], Optional[str], Union[int, str, None], Optional[str], Optional[str]) -> None
        """Waits for PostgreSQL database to be up

        Args:
            database (Optional[str]): Database name
            host (Optional[str]): Host where database is located
            port (Union[int, str, None]): Database port
            username (Optional[str]): Username to log into database
            password (Optional[str]): Password to log into database

        Returns:
            None
        """
        connecting_string = 'Checking for PostgreSQL...'
        if port is not None:
            port = int(port)
        while True:
            try:
                logger.info(connecting_string)
                connection = psycopg2.connect(
                    database=database,
                    host=host,
                    port=port,
                    user=username,
                    password=password,
                    connect_timeout=3
                )
                connection.close()
                logger.info('PostgreSQL is running!')
                break
            except psycopg2.OperationalError:
                time.sleep(1)



# -*- coding: UTF-8 -*-
"""Email Tests"""
from os.path import join

import pytest

from hdx.utilities.email import Email, EmailConfigurationError


class TestEmail:
    @pytest.fixture(scope='class')
    def email_json(self, configfolder):
        return join(configfolder, 'hdx_email_configuration.json')

    @pytest.fixture(scope='class')
    def email_yaml(self, configfolder):
        return join(configfolder, 'hdx_email_configuration.yml')

    def test_mail(self, mocksmtp):
        smtp_initargs = {
            'host': 'localhost',
            'port': 123,
            'local_hostname': 'mycomputer.fqdn.com',
            'timeout': 3,
            'source_address': ('machine', 456),
        }
        username = 'user@user.com'
        password = 'pass'
        email_config_dict = {
            'connection_type': 'ssl',
            'username': username,
            'password': password
        }
        email_config_dict.update(smtp_initargs)

        recipients = ['larry@gmail.com', 'moe@gmail.com', 'curly@gmail.com']
        cc = ['tweedledum@gmail.com', 'tweedledee@gmail.com']
        bcc = ['theinvisibleman@gmail.com', 'ghost@gmail.com']
        subject = 'hello'
        text_body = 'hello there'
        html_body = """\
<html>
  <head></head>
  <body>
    <p>Hi!<br>
       How are you?<br>
       Here is the <a href="https://www.python.org">link</a> you wanted.
    </p>
  </body>
</html>
        """
        sender = 'me@gmail.com'
        mail_options = ['a', 'b']
        rcpt_options = [1, 2]

        with Email(email_config_dict=email_config_dict) as email:
            email.send(recipients, subject, text_body, sender=sender, cc=cc, bcc=bcc, mail_options=mail_options,
                       rcpt_options=rcpt_options)
            assert email.server.type == 'smtpssl'
            assert email.server.initargs == smtp_initargs
            assert email.server.username == username
            assert email.server.password == password
            assert email.server.sender == sender
            assert email.server.recipients == recipients + cc + bcc
            assert email.server.msg == '''Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Subject: hello
From: me@gmail.com
To: larry@gmail.com, moe@gmail.com, curly@gmail.com
Cc: tweedledum@gmail.com, tweedledee@gmail.com

hello there'''
            assert email.server.send_args == {'mail_options': ['a', 'b'], 'rcpt_options': [1, 2]}
            email.send(recipients, subject, text_body, html_body=html_body, sender=sender, mail_options=mail_options,
                       rcpt_options=rcpt_options)
            assert 'Content-Type: multipart/alternative;' in email.server.msg
            assert '''\
MIME-Version: 1.0
Subject: hello
From: me@gmail.com
To: larry@gmail.com, moe@gmail.com, curly@gmail.com''' in email.server.msg
            assert '''\
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

hello there''' in email.server.msg
            assert '''\
Content-Type: text/html; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

<html>
  <head></head>
  <body>
    <p>Hi!<br>
       How are you?<br>
       Here is the <a href="https://www.python.org">link</a> you wanted.
    </p>
  </body>
</html>''' in email.server.msg
            email.send(recipients, subject, text_body, mail_options=mail_options, rcpt_options=rcpt_options)
            assert email.server.sender == username

    def test_json(self, mocksmtp, email_json):
        with Email(email_config_json=email_json) as email:
            email.connect()
            assert email.server.type == 'lmtp'

    def test_yaml(self, mocksmtp, email_yaml):
        with Email(email_config_yaml=email_yaml) as email:
            email.connect()
            assert email.server.type == 'smtp'

    def test_fail(self, mocksmtp, email_json, email_yaml):
        email_config_dict = {
            'connection_type': 'ssl',
        }
        default_email_config_yaml = Email.default_email_config_yaml
        Email.default_email_config_yaml = 'NOT EXIST'
        with pytest.raises(IOError):
            with Email() as email:
                pass
        Email.default_email_config_yaml = default_email_config_yaml
        with pytest.raises(EmailConfigurationError):
            with Email(email_config_dict=email_config_dict, email_config_json=email_json) as email:
                pass
        with pytest.raises(EmailConfigurationError):
            with Email(email_config_dict=email_config_dict, email_config_yaml=email_yaml) as email:
                pass



# -*- coding: utf-8 -*-
"""Text processing utilities"""
import difflib
import logging
import re
from string import punctuation
from typing import List, Dict, Optional

logger = logging.getLogger(__name__)

PUNCTUATION_MINUS_BRACKETS = r"""!"#$%&'*+,-./:;<=>?@\^_`|~"""


def remove_end_characters(string, characters_to_remove=punctuation):
    # type: (str, str) -> str
    """Remove any characters at end of string that are in characters_to_remove

    Args:
        string (str): Input string
        characters_to_remove (str): Characters to remove. Defaults to punctuation.

    Returns:
        str: String with any characters at end of string that are in characters_to_remove removed

    """
    while string[-1] in characters_to_remove:
        string = string[:-1]
    return string


def remove_from_end(string, things_to_remove, logging_text=None, whole_words=True):
    # type: (str, List[str], Optional[str], bool) -> str
    """Remove list of items from end of string, stripping any whitespace

    Args:
        string (str): Input string
        things_to_remove (List[str]): Things to remove from the end of string
        logging_text (Optional[str]): Text to log. Defaults to None.
        whole_words (bool): Remove parts of or whole words. Defaults to True (whole words only).

    Returns:
        str: String with text removed

    """
    for thing in things_to_remove:
        thing_len = len(thing)
        string_len = len(string)
        if string_len <= thing_len + 1:
            continue
        position = -thing_len
        if string[position:] != thing:
            continue
        if whole_words and string[position - 1].isalpha():
            continue
        newstring = string[:position].strip()
        if logging_text:
            logger.info(logging_text % (string, newstring))
        string = newstring
    return string


def remove_string(string, toremove, end_characters_to_remove=punctuation):
    # type: (str, str, str) -> str
    """
    Remove string from another string and delete any preceding end characters - by default punctuation (eg. comma)
    and any whitespace following the punctuation

    Args:
        string (str): String to process
        toremove (str): String to remove
        end_characters_to_remove (str): Characters to remove. Defaults to punctuation.

    Returns:
        str: String with other string removed

    """
    index = string.find(toremove)
    newstring = string[:index].strip()
    newstring = remove_end_characters(newstring, characters_to_remove=end_characters_to_remove)
    return '%s%s' % (newstring, string[index + len(toremove):])


def multiple_replace(string, replacements):
    # type: (str, Dict[str,str]) -> str
    """Simultaneously replace multiple strings in a string

    Args:
        string (str): Input string
        replacements (Dict[str,str]): Replacements dictionary

    Returns:
        str: String with replacements

    """
    pattern = re.compile("|".join([re.escape(k) for k in sorted(replacements, key=len, reverse=True)]), flags=re.DOTALL)
    return pattern.sub(lambda x: replacements[x.group(0)], string)


def get_words_in_sentence(sentence):
    # type: (str) -> List[str]
    """Returns list of words in a sentence

    Args:
        sentence (str): Sentence

    Returns:
        List[str]: List of words in sentence

    """
    return re.sub('[' + punctuation.replace("'", "") + ']', ' ', sentence).split()


def get_matching_text_in_strs(a, b, match_min_size=30, ignore='', end_characters=''):
    # type: (str, str, int, str, str) -> List[str]
    """Returns a list of matching blocks of text in a and b

    Args:
        a (str): First string to match
        b (str): Second string to match
        match_min_size (int): Minimum block size to match on. Defaults to 30.
        ignore (str): Any characters to ignore in matching. Defaults to ''.
        end_characters (str): End characters to look for. Defaults to ''.

    Returns:
        List[str]: List of matching blocks of text

    """
    compare = difflib.SequenceMatcher(lambda x: x in ignore)
    compare.set_seqs(a=a, b=b)
    matching_text = list()

    for match in compare.get_matching_blocks():
        start = match.a
        text = a[start: start+match.size]
        if end_characters:
            prev_text = text
            while len(text) != 0 and text[0] in end_characters:
                text = text[1:]
            while len(text) != 0 and text[-1] not in end_characters:
                text = text[:-1]
            if len(text) == 0:
                text = prev_text
        if len(text) >= match_min_size:
            matching_text.append(text)
    return matching_text


def get_matching_text(string_list, match_min_size=30, ignore='', end_characters='.!\r\n'):
    # type: (List[str], int, str, str) -> str
    """Returns a string containing matching blocks of text in a list of strings followed by non-matching.

    Args:
        string_list (List[str]): List of strings to match
        match_min_size (int): Minimum block size to match on. Defaults to 30.
        ignore (str): Any characters to ignore in matching. Defaults to ''.
        end_characters (str): End characters to look for. Defaults to '.\r\n'.

    Returns:
        str: String containing matching blocks of text followed by non-matching

    """
    a = string_list[0]
    for i in range(1, len(string_list)):
        b = string_list[i]
        result = get_matching_text_in_strs(a, b, match_min_size=match_min_size, ignore=ignore,
                                           end_characters=end_characters)
        a = ''.join(result)
    return a


def get_matching_then_nonmatching_text(string_list, separator='', match_min_size=30, ignore='',
                                       end_characters='.!\r\n'):
    # type: (List[str], str, int, str, str) -> str
    """Returns a string containing matching blocks of text in a list of strings followed by non-matching.

    Args:
        string_list (List[str]): List of strings to match
        separator (str): Separator to add between blocks of text. Defaults to ''.
        match_min_size (int): Minimum block size to match on. Defaults to 30.
        ignore (str): Any characters to ignore in matching. Defaults to ''.
        end_characters (str): End characters to look for. Defaults to '.\r\n'.

    Returns:
        str: String containing matching blocks of text followed by non-matching

    """
    def add_separator_if_needed(text_list):
        if separator and len(text_list) > 0 and text_list[-1][-len(separator):] != separator:
            text_list.append(separator)

    a = string_list[0]
    for i in range(1, len(string_list)):
        b = string_list[i]
        combined_len = len(a) + len(b)
        result = get_matching_text_in_strs(a, b, match_min_size=match_min_size, ignore=ignore,
                                           end_characters=end_characters)
        new_a = a
        new_b = b
        for text in result:
            new_a = new_a.replace(text, '')
            new_b = new_b.replace(text, '')
        if new_a and new_a in a:
            pos_a = a.index(new_a)
        else:
            pos_a = combined_len
        if new_b and new_b in b:
            pos_b = b.index(new_b)
        else:
            pos_b = combined_len
        if pos_b > pos_a:
            text_1 = new_b
            pos_1 = pos_b
            text_2 = new_a
            pos_2 = pos_a
        else:
            text_1 = new_a
            pos_1 = pos_a
            text_2 = new_b
            pos_2 = pos_b
        output = list()
        pos = 0
        for text in result:
            output.append(text)
            pos += len(text)
            if text_1 and pos >= pos_1:
                add_separator_if_needed(output)
                output.append(text_1)
                pos += len(text_1)
                text_1 = None
            if text_2 and pos >= pos_2:
                add_separator_if_needed(output)
                output.append(text_2)
                pos += len(text_2)
                text_2 = None
        if text_1 and pos_1 == combined_len:
            add_separator_if_needed(output)
            output.append(text_1)
        if text_2 and pos_2 == combined_len:
            add_separator_if_needed(output)
            output.append(text_2)
        a = ''.join(output)
    return a



# -*- coding: UTF-8 -*-
"""Dictionary Tests"""
from os import remove
from os.path import join
from tempfile import gettempdir

import pytest

from hdx.utilities.dictandlist import merge_dictionaries, dict_diff, dict_of_lists_add, list_distribute_contents, \
    list_distribute_contents_simple, extract_list_from_list_of_dict, avg_dicts, key_value_convert, integer_key_convert, \
    integer_value_convert, float_value_convert, write_list_to_csv, read_list_from_csv, dict_of_sets_add, args_to_dict


class TestDictAndList:
    def test_merge_dictionaries(self):
        d1 = {1: 1, 2: 2, 3: 3, 4: {'a': 1, 'b': 'c'}}
        d2 = {2: 6, 4: 8, 6: 9, 9: {'c': 12, 'e': 'h'}}
        d3 = {4: {'g': 3}, 5: 7, 8: {3: 12, 'k': 'b'}}
        with pytest.raises(ValueError):
            result = merge_dictionaries([d1, d2, d3])
        d2[4] = {8: '8'}
        result = merge_dictionaries([d1, d2, d3])
        assert result == {1: 1, 2: 6, 3: 3, 4: {8: '8', 'g': 3, 'b': 'c', 'a': 1}, 5: 7, 6: 9, 8: {'k': 'b', 3: 12},
                          9: {'e': 'h', 'c': 12}}
        d1 = {1: 1, 2: 2, 3: 3, 4: ['a', 'b', 'c']}
        d2 = {2: 6, 5: 8, 6: 9, 4: ['d', 'e']}
        result = merge_dictionaries([d1, d2])
        assert result == {1: 1, 2: 6, 3: 3, 4: ['d', 'e'], 5: 8, 6: 9}
        d1 = {1: 1, 2: 2, 3: 3, 4: ['a', 'b', 'c']}
        result = merge_dictionaries([d1, d2], merge_lists=True)
        assert result == {1: 1, 2: 6, 3: 3, 4: ['a', 'b', 'c', 'd', 'e'], 5: 8, 6: 9}

    def test_dict_diff(self):
        d1 = {1: 1, 2: 2, 3: 3, 4: {'a': 1, 'b': 'c'}}
        d2 = {4: {'a': 1, 'b': 'c'}, 2: 2, 3: 3, 1: 1}
        diff = dict_diff(d1, d2)
        assert diff == {}
        d2[3] = 4
        diff = dict_diff(d1, d2)
        assert diff == {3: (3, 4)}
        del d2[3]
        diff = dict_diff(d1, d2)
        assert diff == {3: (3, '<KEYNOTFOUND>')}
        d2[3] = 3
        del d1[4]
        diff = dict_diff(d1, d2)
        assert diff == {4: ('<KEYNOTFOUND>', {'a': 1, 'b': 'c'})}

    def test_dict_of_lists_add(self):
        d = dict()
        dict_of_lists_add(d, 'a', 1)
        assert d == {'a': [1]}
        dict_of_lists_add(d, 2, 'b')
        assert d == {'a': [1], 2: ['b']}
        dict_of_lists_add(d, 'a', 2)
        assert d == {'a': [1, 2], 2: ['b']}
        dict_of_lists_add(d, 2, 'c')
        assert d == {'a': [1, 2], 2: ['b', 'c']}
        dict_of_lists_add(d, 2, 'b')
        assert d == {'a': [1, 2], 2: ['b', 'c', 'b']}

    def test_dict_of_sets_add(self):
        d = dict()
        dict_of_sets_add(d, 'a', 1)
        assert d == {'a': {1}}
        dict_of_sets_add(d, 2, 'b')
        assert d == {'a': {1}, 2: {'b'}}
        dict_of_sets_add(d, 'a', 2)
        assert d == {'a': {1, 2}, 2: {'b'}}
        dict_of_sets_add(d, 2, 'c')
        assert d == {'a': {1, 2}, 2: {'b', 'c'}}
        dict_of_sets_add(d, 2, 'b')
        assert d == {'a': {1, 2}, 2: {'b', 'c'}}

    def test_list_distribute_contents_simple(self):
        input_list = [3, 1, 1, 1, 2, 2]
        result = list_distribute_contents_simple(input_list)
        assert result == [1, 2, 3, 1, 2, 1]
        input_list = [('f', 1), ('d', 7), ('d', 2), ('c', 2), ('c', 5), ('c', 4), ('f', 2)]
        result = list_distribute_contents_simple(input_list, lambda x: x[0])
        assert result == [('c', 2), ('d', 7), ('f', 1), ('c', 5), ('d', 2), ('f', 2), ('c', 4)]
        input_list = [{'key': 'd', 'data': 5}, {'key': 'd', 'data': 1}, {'key': 'g', 'data': 2},
                      {'key': 'a', 'data': 2}, {'key': 'a', 'data': 3}, {'key': 'b', 'data': 5}]
        result = list_distribute_contents_simple(input_list, lambda x: x['key'])
        assert result == [{'key': 'a', 'data': 2}, {'key': 'b', 'data': 5}, {'key': 'd', 'data': 5},
                          {'key': 'g', 'data': 2}, {'key': 'a', 'data': 3}, {'key': 'd', 'data': 1}]

    def test_list_distribute_contents(self):
        input_list = [3, 1, 1, 1, 2, 2]
        result = list_distribute_contents(input_list)
        assert result == [1, 2, 1, 2, 1, 3]
        input_list = [('f', 1), ('d', 7), ('d', 2), ('c', 2), ('c', 5), ('c', 4), ('f', 2)]
        result = list_distribute_contents(input_list, lambda x: x[0])
        assert result == [('c', 2), ('d', 7), ('f', 1), ('c', 5), ('d', 2), ('c', 4), ('f', 2)]
        input_list = [{'key': 'd', 'data': 5}, {'key': 'd', 'data': 1}, {'key': 'g', 'data': 2},
                      {'key': 'a', 'data': 2}, {'key': 'a', 'data': 3}, {'key': 'b', 'data': 5}]
        result = list_distribute_contents(input_list, lambda x: x['key'])
        assert result == [{'key': 'a', 'data': 2}, {'key': 'd', 'data': 5}, {'key': 'b', 'data': 5},
                          {'key': 'a', 'data': 3}, {'key': 'd', 'data': 1}, {'key': 'g', 'data': 2}]

    def test_extract_list_from_list_of_dict(self):
        input_list = [{'key': 'd', 1: 5}, {'key': 'd', 1: 1}, {'key': 'g', 1: 2},
                      {'key': 'a', 1: 2}, {'key': 'a', 1: 3}, {'key': 'b', 1: 5}]
        result = extract_list_from_list_of_dict(input_list, 'key')
        assert result == ['d', 'd', 'g', 'a', 'a', 'b']
        result = extract_list_from_list_of_dict(input_list, 1)
        assert result == [5, 1, 2, 2, 3, 5]

    def test_key_value_convert(self):
        d1 = {1: 1, 2: 1.0, 3: 3, 4: 4}
        assert key_value_convert(d1) == d1
        d1 = {1: 2, 2: 2.0, 3: 5, 'la': 4}
        assert key_value_convert(d1, keyfn=int) == {1: 2, 2: 2.0, 3: 5, 'la': 4}
        assert key_value_convert(d1, keyfn=int, dropfailedkeys=True) == {1: 2, 2: 2.0, 3: 5}
        d1 = {1: 2, 2: 2.0, 3: 5, 4: 'la'}
        assert key_value_convert(d1, valuefn=int) == {1: 2, 2: 2.0, 3: 5, 4: 'la'}
        assert key_value_convert(d1, valuefn=int, dropfailedvalues=True) == {1: 2, 2: 2.0, 3: 5}

    def test_integer_key_convert(self):
        d1 = {1: 1, 2: 1.5, 3.5: 3, '4': 4}
        assert integer_key_convert(d1) == {1: 1, 2: 1.5, 3: 3, 4: 4}

    def test_integer_value_convert(self):
        d1 = {1: 1, 2: 1.5, 3: '3', 4: 4}
        assert integer_value_convert(d1) == {1: 1, 2: 1, 3: 3, 4: 4}

    def test_float_value_convert(self):
        d1 = {1: 1, 2: 1.5, 3: '3', 4: 4}
        assert float_value_convert(d1) == {1: 1.0, 2: 1.5, 3: 3.0, 4: 4.0}

    def test_avg_dicts(self):
        d1 = {1: 1, 2: 1.0, 3: 3, 4: 4}
        d2 = {1: 2, 2: 2.0, 3: 5, 4: 4, 7: 3}
        assert avg_dicts(d1, d2) == {1: 1.5, 2: 1.5, 3: 4, 4: 4}
        assert avg_dicts(d1, d2, dropmissing=False) == {1: 1.5, 2: 1.5, 3: 4, 4: 4, 7: 3}

    def test_read_write_list_to_csv(self):
        l = [[1, 2, 3, 'a'],
             [4, 5, 6, 'b'],
             [7, 8, 9, 'c']]
        folder = gettempdir()
        filename = 'test_read_write_list_to_csv.csv'
        filepath = join(folder, filename)
        write_list_to_csv(filepath, l, headers=['h1', 'h2', 'h3', 'h4'])
        newll = read_list_from_csv(filepath)
        newld = read_list_from_csv(filepath, headers=1, dict_form=True)
        remove(filepath)
        assert newll == [['h1', 'h2', 'h3', 'h4'], ['1', '2', '3', 'a'], ['4', '5', '6', 'b'], ['7', '8', '9', 'c']]
        assert newld == [{'h1': '1', 'h2': '2', 'h4': 'a', 'h3': '3'},
                         {'h1': '4', 'h2': '5', 'h4': 'b', 'h3': '6'},
                         {'h1': '7', 'h2': '8', 'h4': 'c', 'h3': '9'}]
        with pytest.raises(ValueError):
            read_list_from_csv(filepath, dict_form=True)

    def test_args_to_dict(self):
        args = 'a=1,big=hello,1=3'
        assert args_to_dict(args) == {'a': '1', 'big': 'hello', '1': '3'}



# -*- coding: UTF-8 -*-
"""Global fixtures"""
import smtplib
from os.path import join

import pytest


@pytest.fixture(scope='session')
def fixturesfolder():
    return join('tests', 'fixtures')


@pytest.fixture(scope='session')
def configfolder(fixturesfolder):
        return join(fixturesfolder, 'config')


@pytest.fixture(scope='function')
def mocksmtp(monkeypatch):
    class MockSMTPBase(object):
        type = None

        def __init__(self, **kwargs):
            self.initargs = kwargs

        def login(self, username, password):
            self.username = username
            self.password = password

        def sendmail(self, sender, recipients, msg, **kwargs):
            self.sender = sender
            self.recipients = recipients
            self.msg = msg
            self.send_args = kwargs

        @staticmethod
        def quit():
            pass

    class MockSMTPSSL(MockSMTPBase):
        type = 'smtpssl'

    class MockLMTP(MockSMTPBase):
        type = 'lmtp'

    class MockSMTP(MockSMTPBase):
        type = 'smtp'
    monkeypatch.setattr(smtplib, 'SMTP_SSL', MockSMTPSSL)
    monkeypatch.setattr(smtplib, 'LMTP', MockLMTP)
    monkeypatch.setattr(smtplib, 'SMTP', MockSMTP)



# -*- coding: utf-8 -*-
"""Directory Path Utilities"""
import contextlib
import inspect
import logging
import sys
from os import getenv, makedirs
from os.path import abspath, realpath, dirname, join, exists
from shutil import rmtree
from tempfile import gettempdir
from typing import Any, Optional, Iterable, Tuple, Dict

from hdx.utilities.loader import load_file_to_str
from hdx.utilities.saver import save_str_to_file

logger = logging.getLogger(__name__)


def script_dir(pyobject, follow_symlinks=True):
    # type: (Any, bool) -> str
    """Get current script's directory

    Args:
        pyobject (Any): Any Python object in the script
        follow_symlinks (bool): Follow symlinks or not. Defaults to True.

    Returns:
        str: Current script's directory
    """
    if getattr(sys, 'frozen', False):  # py2exe, PyInstaller, cx_Freeze
        path = abspath(sys.executable)   # pragma: no cover
    else:
        path = inspect.getabsfile(pyobject)
    if follow_symlinks:
        path = realpath(path)
    return dirname(path)


def script_dir_plus_file(filename, pyobject, follow_symlinks=True):
    # type: (str, Any, bool) -> str
    """Get current script's directory and then append a filename

    Args:
        filename (str): Filename to append to directory path
        pyobject (Any): Any Python object in the script
        follow_symlinks (bool): Follow symlinks or not. Defaults to True.

    Returns:
        str: Current script's directory and with filename appended
    """
    return join(script_dir(pyobject, follow_symlinks), filename)


def get_temp_dir(folder=None):
    # type: (Optional[str]) -> str
    """Get a temporary directory. Looks for environment variable TEMP_DIR and falls
    back on os.gettempdir.

    Args:
        folder (Optional[str]): Folder to create in temporary folder. Defaults to None.

    Returns:
        str: A temporary directory
    """
    tempdir = getenv('TEMP_DIR', gettempdir())
    if folder:
        tempdir = join(tempdir, folder)
        if not exists(tempdir):
            makedirs(tempdir)
    return tempdir


@contextlib.contextmanager
def temp_dir(folder=None, delete_on_success=True, delete_on_failure=True):
    # type: (Optional[str], bool, bool) -> str
    """Get a temporary directory optionally with folder appended (and created if it doesn't exist)

    Args:
        folder (Optional[str]): Folder to create in temporary folder. Defaults to None.
        delete_on_success (bool): Whether to delete folder (if folder supplied) on exiting with statement successfully. Defaults to True.
        delete_on_failure (bool): Whether to delete folder (if folder supplied) on exiting with statement unsuccessfully. Defaults to True.

    Returns:
        str: A temporary directory
    """
    tempdir = get_temp_dir(folder)
    try:
        yield tempdir
        if folder and delete_on_success:
            rmtree(tempdir)
    except:
        if folder and delete_on_failure:
            rmtree(tempdir)
        raise


def progress_storing_tempdir(folder, iterator, key):
    # type: (str, Iterable[Dict], str) -> Tuple[str,Dict]
    """Yield a temporary directory optionally with folder appended (and created if it doesn't exist)
    and the next dictionary in the iterator. The folder persists until the final iteration and which
    iteration to start at is persisted between runs.

    Args:
        folder (str): Folder to create in temporary folder
        iterator (Iterable[Dict]): Iterate over this object persisting progress
        key (str): Key to examine from dictionary from iterator

    Returns:
        Tuple[str,Dict]: A tuple of the form (temporary directory, next object in iterator)
    """
    with temp_dir(folder, delete_on_success=True, delete_on_failure=False) as tempdir:
        progress_file = join(tempdir, 'progress.txt')
        wheretostart = getenv('WHERETOSTART')
        if wheretostart:
            wheretostart = wheretostart.upper()
            if wheretostart == 'RESET':
                rmtree(tempdir)
                makedirs(tempdir)
                wheretostart = None
                logger.info('Removing progress file and will start from beginning!')
            else:
                logger.info('Environment variable WHERETOSTART = %s' % wheretostart)
        else:
            if exists(progress_file):
                wheretostart = load_file_to_str(progress_file, strip=True)
                logger.info('File WHERETOSTART = %s' % wheretostart)
        found = False
        for nextdict in iterator:
            currentlocation = nextdict[key]
            if wheretostart and not found:
                if currentlocation == wheretostart:
                    found = True
                    logger.info('Starting run from WHERETOSTART %s' % wheretostart)
                else:
                    logger.info('Run not started. Ignoring %s. WHERETOSTART (%s) not matched.' % (currentlocation,
                                                                                                  wheretostart))
                    continue
            save_str_to_file(currentlocation, progress_file)
            yield tempdir, nextdict



import os
import sys
from distutils import log
from distutils.command.clean import clean
from os.path import exists
from shutil import rmtree
from typing import Any
from uuid import UUID

import six
from setuptools import Command


def raisefrom(exc_type, message, exc):
    # type: (Any, str, BaseException) -> None
    """Call Python 3 raise from or emulate it for Python 2

    Args:
        exc_type (Any): Type of Exception
        message (str): Error message to display
        exc (BaseException): original exception

    Returns:
        None

    """
    if sys.version_info[:2] >= (3, 2):
        six.raise_from(exc_type(message), exc)
    else:
        six.reraise(exc_type, '%s - %s' % (message, exc), sys.exc_info()[2])


def is_valid_uuid(uuid_to_test, version=4):
    # type: (str, int) -> bool
    """
    Check if uuid_to_test is a valid UUID.

    Args:
        uuid_to_test (str): UUID to test for validity
        version (int): UUID version. Defaults to 4.

    Returns:
        str: Current script's directory
    """
    try:
        uuid_obj = UUID(uuid_to_test, version=version)
    except:
        return False
    return str(uuid_obj) == uuid_to_test


class CleanCommand(clean):
    """
    Custom implementation of ``clean`` setuptools command."""

    def run(self):  # pragma: no cover
        """After calling the super class implementation, this function removes
        the dist directory if it exists."""
        self.all = True  # --all by default when cleaning
        super(CleanCommand, self).run()
        dir_ = 'dist'
        if exists(dir_):
            log.info("removing '%s' (and everything under it)", dir_)
            rmtree(dir_)
        else:
            log.info("'%s' does not exist -- can't clean it", dir_)


class PackageCommand(Command):
    """Package command for setup.py that creates source and wheel packages."""

    description = 'Build the packages.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        log.info('Building Source and Wheel (universal) Packages...')
        os.system('{0} setup.py clean sdist bdist_wheel --universal'.format(sys.executable))
        sys.exit()


class PublishCommand(Command):
    """Publish command for setup.py that creates git tags and publishes to pypi.
    Requires that twine and git be installed."""

    version = None
    description = 'Publish the packages.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        log.info('Uploading the package to PyPI using twine...')
        os.system('twine upload dist/*')

        if PublishCommand.version:
            log.info('Pushing git tags...')
            os.system('git tag v{0}'.format(PublishCommand.version))
            os.system('git push --tags')

        sys.exit()
